import asyncio
import logging
import warnings

from time import time
from .. import bhf_signals
from contextlib import suppress
from inspect import isawaitable
from ipaddress import ip_address
from typing import Callable, Dict, Generator, Optional, Tuple, Type, TypeVar

from playwright.async_api import (
    BrowserContext,
    Page,
    PlaywrightContextManager,
    Request as PlaywrightRequest,
    Response as PlaywrightResponse,
    Route,
)
from scrapy import Spider, signals
from scrapy.crawler import Crawler
from scrapy.http import Request, Response
from scrapy.http.headers import Headers
from scrapy.utils.misc import load_object
from scrapy.utils.python import to_unicode
from scrapy.responsetypes import responsetypes
from scrapy.utils.defer import deferred_from_coro
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.utils.reactor import verify_installed_reactor
from twisted.internet.defer import Deferred, inlineCallbacks
from scrapy.core.downloader.handlers.http import HTTPDownloadHandler
from w3lib.encoding import html_body_declared_encoding, http_content_type_encoding

from .headers import use_scrapy_headers
from .page import PageMethod

__all__ = ["ScrapyPlaywrightDownloadHandler"]

PlaywrightHandler = TypeVar("PlaywrightHandler", bound="ScrapyPlaywrightDownloadHandler")

logger = logging.getLogger("scrapy-playwright")


def _make_request_logger(context_name: str) -> Callable:
    def _log_request(request: PlaywrightRequest) -> None:
        logger.debug(
            f"[Context={context_name}] Request: <{request.method.upper()} {request.url}> "
            f"(resource type: {request.resource_type}, referrer: {request.headers.get('referer')})"
        )

    return _log_request


def _make_response_logger(context_name: str) -> Callable:
    def _log_request(response: PlaywrightResponse) -> None:
        logger.debug(
            f"[Context={context_name}] Response: <{response.status} {response.url}> "
            f"(referrer: {response.headers.get('referer')})"
        )

    return _log_request


class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
    def __init__(self, crawler: Crawler) -> None:
        super().__init__(settings=crawler.settings, crawler=crawler)
        verify_installed_reactor("twisted.internet.asyncioreactor.AsyncioSelectorReactor")
        crawler.signals.connect(self._engine_started, signals.engine_started)
        self.stats = crawler.stats

        self.browser_type: str = crawler.settings.get("PLAYWRIGHT_BROWSER_TYPE") or "chromium"
        self.max_pages_per_context: int = crawler.settings.getint(
            "PLAYWRIGHT_MAX_PAGES_PER_CONTEXT"
        ) or crawler.settings.getint("CONCURRENT_REQUESTS")
        self.launch_options: dict = crawler.settings.getdict("PLAYWRIGHT_LAUNCH_OPTIONS") or {}

        self.default_navigation_timeout: Optional[float] = None
        if "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT" in crawler.settings:
            with suppress(TypeError, ValueError):
                self.default_navigation_timeout = float(
                    crawler.settings.get("PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT")
                )

        if crawler.settings.get("PLAYWRIGHT_PROCESS_REQUEST_HEADERS"):
            self.process_request_headers = load_object(
                crawler.settings["PLAYWRIGHT_PROCESS_REQUEST_HEADERS"]
            )
        else:
            self.process_request_headers = use_scrapy_headers

        self.context_kwargs: dict = crawler.settings.getdict("PLAYWRIGHT_CONTEXTS")
        self.contexts: Dict[str, BrowserContext] = {}
        self.context_semaphores: Dict[str, asyncio.Semaphore] = {
            'default': asyncio.Semaphore(value=self.max_pages_per_context)
        }

        self.abort_request: Optional[Callable[[PlaywrightRequest], bool]] = None
        if crawler.settings.get("PLAYWRIGHT_ABORT_REQUEST"):
            self.abort_request = load_object(crawler.settings["PLAYWRIGHT_ABORT_REQUEST"])

        self.user_agent_rotation_idx = 0
        self.playwright_user_agents: list = crawler.settings.get("PLAYWRIGHT_USER_AGENTS") or []
        crawler.signals.connect(self._ip_changed, bhf_signals.ip_changed)

    @classmethod
    def from_crawler(cls: Type[PlaywrightHandler], crawler: Crawler) -> PlaywrightHandler:
        return cls(crawler)

    def _engine_started(self) -> Deferred:
        """Launch the browser. Use the engine_started signal as it supports returning deferreds."""
        return deferred_from_coro(self._launch_browser())

    def _ip_changed(self) -> Deferred:
        """Plan User-Agent rotation"""
        return deferred_from_coro(self._plan_user_agent_rotation())

    async def _launch_browser(self, user_agent=None) -> None:
        self.playwright_context_manager = PlaywrightContextManager()
        self.playwright = await self.playwright_context_manager.start()
        logger.info("Launching browser")
        browser_launcher = getattr(self.playwright, self.browser_type).launch
        self.browser = await browser_launcher(**self.launch_options)
        logger.info(f"Browser {self.browser_type} launched")
        context_kwargs = self.context_kwargs
        if user_agent is not None:
            context_kwargs['user_agent'] = user_agent
        self.contexts = {
            'default': await self._create_browser_context('default', context_kwargs)
        }

    async def _create_browser_context(self, name: str, context_kwargs: dict) -> BrowserContext:
        context = await self.browser.new_context(**context_kwargs)
        context.on("close", self._make_close_browser_context_callback(name))
        logger.debug("Browser context started: '%s'", name)
        self.stats.inc_value("playwright/context_count")
        if self.default_navigation_timeout is not None:
            context.set_default_navigation_timeout(self.default_navigation_timeout)
        return context

    async def _plan_user_agent_rotation(self) -> None:
        logger.info("Re create browser: Before acquire")
        await self.context_semaphores['default'].acquire()
        logger.info("Re create browser: After acquire")

        await self._close()
        if not self.playwright_user_agents:
            await self._launch_browser()
        else:
            logger.info(f"Re create browser with new user-agent: "
                        f"{self.playwright_user_agents[self.user_agent_rotation_idx]}")
            await self._launch_browser(user_agent=self.playwright_user_agents[self.user_agent_rotation_idx])
            self.user_agent_rotation_idx = self.user_agent_rotation_idx + 1
            if self.user_agent_rotation_idx >= len(self.playwright_user_agents):
                self.user_agent_rotation_idx = 0

        self.context_semaphores['default'].release()
        logger.info("Re create browser: Completed")

    async def _create_page(self, request: Request) -> Page:
        """Create a new page in a context, also creating a new context if necessary."""
        context = self.contexts.get("default")

        page = await context.new_page()
        self.stats.inc_value("playwright/page_count")
        logger.debug(
            "[Context=%s] New page created, page count is %i (%i for all contexts)",
            "default",
            len(context.pages),
            self._get_total_page_count(),
        )
        if self.default_navigation_timeout is not None:
            page.set_default_navigation_timeout(self.default_navigation_timeout)

        # page.on("close", self._make_close_page_callback("default"))
        page.on("crash", self._make_close_page_callback("default"))
        page.on("request", _make_request_logger("default"))
        page.on("response", _make_response_logger("default"))
        page.on("request", self._increment_request_stats)
        page.on("response", self._increment_response_stats)

        return page

    def _get_total_page_count(self):
        count = sum([len(context.pages) for context in self.contexts.values()])
        current_max_count = self.stats.get_value("playwright/page_count/max_concurrent")
        if current_max_count is None or count > current_max_count:
            self.stats.set_value("playwright/page_count/max_concurrent", count)
        return count

    @inlineCallbacks
    def close(self) -> Deferred:
        yield super().close()
        yield deferred_from_coro(self._close())
        self.context_semaphores.clear()

    async def _close(self) -> None:
        self.contexts.clear()
        if getattr(self, "browser", None):
            logger.info("Closing browser")
            await self.browser.close()
        await self.playwright_context_manager.__aexit__()

    def download_request(self, request: Request, spider: Spider) -> Deferred:
        return deferred_from_coro(self._download_request(request, spider))

    async def _download_request(self, request: Request, spider: Spider) -> Response:
        page = None
        try:
            logger.info(f"[{request.url}] Before acquire")
            await self.context_semaphores['default'].acquire()
            logger.info(f"[{request.url}] After acquire")
            page = await self._create_page(request)

            # attach event handlers
            event_handlers = request.meta.get("playwright_page_event_handlers") or {}
            for event, handler in event_handlers.items():
                if callable(handler):
                    page.on(event, handler)
                elif isinstance(handler, str):
                    try:
                        page.on(event, getattr(spider, handler))
                    except AttributeError:
                        logger.warning(
                            f"Spider '{spider.name}' does not have a '{handler}' attribute,"
                            f" ignoring handler for event '{event}'"
                        )

            await page.unroute("**")
            await page.route(
                "**",
                self._make_request_handler(
                    method=request.method,
                    scrapy_headers=request.headers,
                    body=request.body,
                    encoding=getattr(request, "encoding", None),
                ),
            )

            result = await self._download_request_with_page(request, page)
        except:
            logger.exception(f"[{request.url}] Error while downloading request")
            return Response('', status=504)
        else:
            logger.info(f"[{request.url}] Return result")
            return result
        finally:
            if page is not None:
                if not page.is_closed():
                    await page.close()
                    self.context_semaphores['default'].release()

    async def _download_request_with_page(self, request: Request, page: Page) -> Response:
        start_time = time()
        response = await page.goto(request.url, wait_until="domcontentloaded")

        logger.info(f"[{request.url}] After goto")
        await self._apply_page_methods(page, request)

        logger.info(f"[{request.url}] After apply page methods")
        body_str = await page.content()
        request.meta["download_latency"] = time() - start_time

        server_ip_address = None
        with suppress(AttributeError, KeyError, ValueError):
            server_addr = await response.server_addr()
            if server_addr is not None:
                server_ip_address = ip_address(server_addr.ipAddress)

        with suppress(AttributeError):
            request.meta["playwright_security_details"] = await response.security_details()

        headers = Headers(response.headers)
        headers.pop("Content-Encoding", None)
        body, encoding = _encode_body(headers=headers, text=body_str)
        respcls = responsetypes.from_args(headers=headers, url=page.url, body=body)
        return respcls(
            url=page.url,
            status=response.status,
            headers=headers,
            body=body,
            request=request,
            flags=["playwright"],
            encoding=encoding,
            ip_address=server_ip_address,
        )

    async def _apply_page_methods(self, page: Page, request: Request) -> None:
        page_methods = request.meta.get("playwright_page_methods") or ()

        if not page_methods and "playwright_page_coroutines" in request.meta:
            page_methods = request.meta["playwright_page_coroutines"]
            warnings.warn(
                "The 'playwright_page_coroutines' request meta key is deprecated,"
                " please use 'playwright_page_methods' instead.",
                category=ScrapyDeprecationWarning,
                stacklevel=1,
            )

        if isinstance(page_methods, dict):
            page_methods = page_methods.values()
        for pm in page_methods:
            if isinstance(pm, PageMethod):
                try:
                    method = getattr(page, pm.method)
                except AttributeError:
                    logger.warning(f"Ignoring {repr(pm)}: could not find method")
                else:
                    result = method(*pm.args, **pm.kwargs)
                    pm.result = await result if isawaitable(result) else result
                    await page.wait_for_load_state(timeout=self.default_navigation_timeout)
            else:
                logger.warning(f"Ignoring {repr(pm)}: expected PageMethod, got {repr(type(pm))}")

    def _increment_request_stats(self, request: PlaywrightRequest) -> None:
        stats_prefix = "playwright/request_count"
        self.stats.inc_value(stats_prefix)
        self.stats.inc_value(f"{stats_prefix}/resource_type/{request.resource_type}")
        self.stats.inc_value(f"{stats_prefix}/method/{request.method}")
        if request.is_navigation_request():
            self.stats.inc_value(f"{stats_prefix}/navigation")

    def _increment_response_stats(self, response: PlaywrightResponse) -> None:
        stats_prefix = "playwright/response_count"
        self.stats.inc_value(stats_prefix)
        self.stats.inc_value(f"{stats_prefix}/resource_type/{response.request.resource_type}")
        self.stats.inc_value(f"{stats_prefix}/method/{response.request.method}")

    def _make_close_page_callback(self, context_name: str) -> Callable:
        def close_page_callback() -> None:
            logger.info("Close page callback")
            if context_name in self.context_semaphores:
                self.context_semaphores[context_name].release()

        return close_page_callback

    def _make_close_browser_context_callback(self, name: str) -> Callable:
        def close_browser_context_callback() -> None:
            logger.debug("Browser context closed: '%s'", name)
            if name in self.contexts:
                self.contexts.pop(name)

        return close_browser_context_callback

    def _make_request_handler(
            self, method: str, scrapy_headers: Headers, body: Optional[bytes], encoding: str = "utf8"
    ) -> Callable:
        async def _request_handler(route: Route, playwright_request: PlaywrightRequest) -> None:
            """Override request headers, method and body."""
            if self.abort_request and self.abort_request(playwright_request):
                await route.abort()
                self.stats.inc_value("playwright/request_count/aborted")
                return None

            processed_headers = await self.process_request_headers(
                self.browser_type, playwright_request, scrapy_headers
            )

            # the request that reaches the callback should contain the headers that were sent
            scrapy_headers.clear()
            scrapy_headers.update(processed_headers)

            overrides: dict = {"headers": processed_headers}
            if playwright_request.is_navigation_request():
                overrides["method"] = method
                if body is not None:
                    overrides["post_data"] = body.decode(encoding)

            await route.continue_(**overrides)

        return _request_handler


def _possible_encodings(headers: Headers, text: str) -> Generator[str, None, None]:
    if headers.get("content-type"):
        content_type = to_unicode(headers["content-type"])
        yield http_content_type_encoding(content_type)
    yield html_body_declared_encoding(text)


def _encode_body(headers: Headers, text: str) -> Tuple[bytes, str]:
    for encoding in filter(None, _possible_encodings(headers, text)):
        try:
            body = text.encode(encoding)
        except UnicodeEncodeError:
            pass
        else:
            return body, encoding
    return text.encode("utf-8"), "utf-8"  # fallback



import mock
from mock import MagicMock
from unittest import TestCase
from scrapy.http import Response
from crawling.redis_stats_middleware import RedisStatsMiddleware

class TestRedisStatsMiddleware(TestCase):

    @mock.patch('crawling.redis_stats_middleware' \
                '.RedisStatsMiddleware.setup')
    def setUp(self, s):
        self.rsm = RedisStatsMiddleware(MagicMock())
        self.rsm.logger = MagicMock()
        self.rsm.logger.debug = MagicMock()
        self.rsm.redis_conn = MagicMock()
        self.rsm.stats_dict = {}
        self.rsm.settings = {}
        self.rsm.settings['STATS_CYCLE'] = 5
        self.rsm.settings['STATS_TIMES'] = []
        self.rsm.settings['STATS_RESPONSE_CODES'] = []
        self.rsm._get_hostname = MagicMock(return_value='host1')

    def test_load_stats_codes(self):
        # test nothing
        spider_name = 'link'
        self.rsm._setup_stats_status_codes(spider_name)
        self.assertEqual(list(self.rsm.stats_dict[spider_name]['status_codes'].keys()), [])

        # test status codes only
        self.rsm.settings['STATS_RESPONSE_CODES'] = [200, 403]
        self.rsm._setup_stats_status_codes(spider_name)
        self.assertEqual(
            sorted(self.rsm.stats_dict[spider_name]['status_codes'].keys()),
            sorted([200, 403]))
        self.assertEqual(list(self.rsm.stats_dict[spider_name]['status_codes'][200].keys()),
                         ['lifetime'])
        self.assertEqual(list(self.rsm.stats_dict[spider_name]['status_codes'][403].keys()),
                         ['lifetime'])

        # test good/bad rolling stats
        self.rsm.stats_dict = {}
        self.rsm.settings['STATS_TIMES'] = [
            'SECONDS_15_MINUTE',
            'SECONDS_1_HOUR',
            'SECONDS_DUMB',
        ]
        good = [
            'lifetime',  # for totals, not DUMB
            '900',
            '3600',
        ]

        # check that both keys are set up
        self.rsm._setup_stats_status_codes(spider_name)
        self.assertEqual(
            sorted([str(x) for x in self.rsm.stats_dict[spider_name]['status_codes'][200].keys()]),
            sorted(good))
        self.assertEqual(
            sorted([str(x) for x in self.rsm.stats_dict[spider_name]['status_codes'][403].keys()]),
            sorted(good))

        k1 = 'stats:crawler:host1:link:200'
        k2 = 'stats:crawler:host1:link:403'

        for time_key in self.rsm.stats_dict[spider_name]['status_codes'][200]:
            if time_key == 0:
                self.assertEqual(
                    self.rsm.stats_dict[spider_name]['status_codes'][200][0].key,
                    '{k}:lifetime'.format(k=k1)
                    )
            else:
                self.assertEqual(
                    self.rsm.stats_dict[spider_name]['status_codes'][200][time_key].key,
                    '{k}:{t}'.format(k=k1, t=time_key)
                    )

        for time_key in self.rsm.stats_dict[spider_name]['status_codes'][403]:
            if time_key == 0:
                self.assertEqual(
                    self.rsm.stats_dict[spider_name]['status_codes'][403][0].key,
                    '{k}:lifetime'.format(k=k2)
                    )
            else:
                self.assertEqual(
                    self.rsm.stats_dict[spider_name]['status_codes'][403][time_key].key,
                    '{k}:{t}'.format(k=k2, t=time_key)
                    )

    def test_rsm_input(self):
        responses = [
            Response('http://istresearch.com', status=200),
            Response('http://istresearch.com', status=400),
            Response('http://istresearch.com', status=404),
            Response('http://istresearch.com', status=200),
            Response('http://istresearch.com', status=999),
        ]

        self.rsm.settings['STATS_TIMES'] = [
            'lifetime'
        ]
        self.rsm.settings['STATS_RESPONSE_CODES'] = [200, 400, 404]
        self.rsm.settings['STATS_STATUS_CODES'] = True

        fake_stats = MagicMock()
        fake_stats.increment = MagicMock()

        self.rsm.stats_dict = {
            'link': {
                'status_codes': {
                    200: {
                        'lifetime': fake_stats
                    },
                    400: {
                        'lifetime': fake_stats
                    },
                    404: {
                        'lifetime': fake_stats
                    }
                }
            },
            'wandering': {
                'status_codes': {
                    200: {
                        'lifetime': fake_stats
                    },
                    400: {
                        'lifetime': fake_stats
                    },
                    404: {
                        'lifetime': fake_stats
                    }
                }
            }
        }

        for name in ['link', 'wandering']:
            spider = MagicMock()
            spider.name = name
            for response in responses:
                self.rsm.process_spider_input(response, spider)

        # 4 calls for link, 4 calls for wandering
        self.assertEqual(fake_stats.increment.call_count, 8)





from __future__ import absolute_import
# This file houses all default settings for the Crawler
# to override please use a custom localsettings.py file

# Scrapy Cluster Settings
# ~~~~~~~~~~~~~~~~~~~~~~~

# Specify the host, port and password to use when connecting to Redis.
REDIS_HOST = 'localhost'
REDIS_PORT = '6379'
REDIS_DB = 0
REDIS_PASSWORD = None
REDIS_SOCKET_TIMEOUT = 10

# Kafka server information
KAFKA_HOSTS = ['localhost:9092']
KAFKA_TOPIC_PREFIX = 'demo'
KAFKA_APPID_TOPICS = False
# base64 encode the html body to avoid json dump errors due to malformed text
KAFKA_BASE_64_ENCODE = False
KAFKA_PRODUCER_BATCH_LINGER_MS = 25  # 25 ms before flush
KAFKA_PRODUCER_BUFFER_BYTES = 4 * 1024 * 1024  # 4MB before blocking
KAFKA_PRODUCER_MAX_REQUEST_SIZE = 1024 * 1024 # 1MB

ZOOKEEPER_ASSIGN_PATH = '/scrapy-cluster/crawler/'
ZOOKEEPER_ID = 'all'
ZOOKEEPER_HOSTS = 'localhost:2181'

PUBLIC_IP_URL = 'http://ip.42.pl/raw'
IP_ADDR_REGEX = '(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'

# Don't cleanup redis queues, allows to pause/resume crawls.
SCHEDULER_PERSIST = True

# seconds to wait between seeing new queues, cannot be faster than spider_idle time of 5
SCHEDULER_QUEUE_REFRESH = 10

# throttled queue defaults per domain, x hits in a y second window
QUEUE_HITS = 10
QUEUE_WINDOW = 60

# we want the queue to produce a consistent pop flow
QUEUE_MODERATED = True

# how long we want the duplicate timeout queues to stick around in seconds
DUPEFILTER_TIMEOUT = 600

# how many pages to crawl for an individual domain. Cluster wide hard limit.
GLOBAL_PAGE_PER_DOMAIN_LIMIT = None

# how long should the global page limit per domain stick around in seconds
GLOBAL_PAGE_PER_DOMAIN_LIMIT_TIMEOUT = 600

# how long should the individual domain's max page limit stick around in seconds
DOMAIN_MAX_PAGE_TIMEOUT = 600

# how often to refresh the ip address of the scheduler
SCHEDULER_IP_REFRESH = 60

# whether to add depth >= 1 blacklisted domain requests back to the queue
SCHEDULER_BACKLOG_BLACKLIST = True

'''
----------------------------------------
The below parameters configure how spiders throttle themselves across the cluster
All throttling is based on the TLD of the page you are requesting, plus any of the
following parameters:

Type: You have different spider types and want to limit how often a given type of
spider hits a domain

IP: Your crawlers are spread across different IP's, and you want each IP crawler clump
to throttle themselves for a given domain

Combinations for any given Top Level Domain:
None - all spider types and all crawler ips throttle themselves from one tld queue
Type only - all spiders throttle themselves based off of their own type-based tld queue,
    regardless of crawler ip address
IP only - all spiders throttle themselves based off of their public ip address, regardless
    of spider type
Type and IP - every spider's throttle queue is determined by the spider type AND the
    ip address, allowing the most fined grained control over the throttling mechanism
'''
# add Spider type to throttle mechanism
SCHEDULER_TYPE_ENABLED = True

# add ip address to throttle mechanism
SCHEDULER_IP_ENABLED = True
'''
----------------------------------------
'''

# how many times to retry getting an item from the queue before the spider is considered idle
SCHEUDLER_ITEM_RETRIES = 3

# how long to keep around stagnant domain queues
SCHEDULER_QUEUE_TIMEOUT = 3600

# log setup scrapy cluster crawler
SC_LOGGER_NAME = 'sc-crawler'
SC_LOG_DIR = 'logs'
SC_LOG_FILE = 'sc_crawler.log'
SC_LOG_MAX_BYTES = 10 * 1024 * 1024
SC_LOG_BACKUPS = 5
SC_LOG_STDOUT = True
SC_LOG_JSON = False
SC_LOG_LEVEL = 'INFO'


# stats setup
STATS_STATUS_CODES = True
STATS_RESPONSE_CODES = [
    200,
    404,
    403,
    504,
]
STATS_CYCLE = 5
# from time variables in scutils.stats_collector class
STATS_TIMES = [
    'SECONDS_15_MINUTE',
    'SECONDS_1_HOUR',
    'SECONDS_6_HOUR',
    'SECONDS_12_HOUR',
    'SECONDS_1_DAY',
    'SECONDS_1_WEEK',
]

# Scrapy Settings
# ~~~~~~~~~~~~~~~
# Scrapy settings for distributed_crawling project
#
BOT_NAME = 'crawling'

SPIDER_MODULES = ['crawling.spiders']
NEWSPIDER_MODULE = 'crawling.spiders'

# Enables scheduling storing requests queue in redis.
SCHEDULER = "crawling.distributed_scheduler.DistributedScheduler"

# Store scraped item in redis for post-processing.
ITEM_PIPELINES = {
    'crawling.pipelines.KafkaPipeline': 100,
    'crawling.pipelines.LoggingBeforePipeline': 1,
}

SPIDER_MIDDLEWARES = {
    # disable built-in DepthMiddleware, since we do our own
    # depth management per crawl request
    'scrapy.spidermiddlewares.depth.DepthMiddleware': None,
    'crawling.meta_passthrough_middleware.MetaPassthroughMiddleware': 100,
    'crawling.redis_stats_middleware.RedisStatsMiddleware': 101
}

DOWNLOADER_MIDDLEWARES = {
    # Handle timeout retries with the redis scheduler and logger
    'scrapy.downloadermiddlewares.retry.RetryMiddleware': None,
    'crawling.redis_retry_middleware.RedisRetryMiddleware': 510,
    # exceptions processed in reverse order
    'crawling.log_retry_middleware.LogRetryMiddleware': 520,
    # custom cookies to not persist across crawl requests
    'scrapy.downloadermiddlewares.cookies.CookiesMiddleware': None,
    'crawling.custom_cookies.CustomCookiesMiddleware': 700,
}

# Disable the built in logging in production
LOG_ENABLED = False

# Allow all return codes
HTTPERROR_ALLOW_ALL = True

RETRY_TIMES = 3

DOWNLOAD_TIMEOUT = 10

# Avoid in-memory DNS cache. See Advanced topics of docs for info
DNSCACHE_ENABLED = True

# Local Overrides
# ~~~~~~~~~~~~~~~

try:
    from .localsettings import *
except ImportError:
    pass



from scrapy.dupefilters import BaseDupeFilter
from scrapy.utils.request import request_fingerprint


class RFPDupeFilter(BaseDupeFilter):
    '''
    Redis-based request duplication filter
    '''

    def __init__(self, server, key, timeout):
        '''
        Initialize duplication filter

        @param server: the redis connection
        @param key: the key to store the fingerprints
        @param timeout: number of seconds a given key will remain once idle
        '''
        self.server = server
        self.key = key
        self.timeout = timeout

    def request_seen(self, request):
        fp = request_fingerprint(request)
        c_id = request.meta['crawlid']

        added = self.server.sadd(self.key + ":" + c_id, fp)
        self.server.expire(self.key + ":" + c_id, self.timeout)

        return not added

    def close(self, reason):
        '''
        Delete data on close. Called by scrapy's scheduler
        '''
        self.clear()

    def clear(self):
        '''
        Clears fingerprints data
        '''
        self.server.delete(self.key)



from __future__ import division

import time
from builtins import str
from builtins import object
from past.utils import old_div

from redis.exceptions import WatchError


class RedisThrottledQueue(object):
    queue = None  # the instantiated queue class
    window = None  # the window to use to limit requests
    limit = None  # number of requests in the given window
    redis_conn = None  # the redis connection
    moderation = None  # whether to use moderation or not
    moderate_key = None  # the last time the moderated queue was pulled
    window_append = ":throttle_window"  # appended to end of window queue key
    time_append = ":throttle_time"  # appended to end to time key
    elastic = False  # use elastic catch up
    elastic_buffer = 0  # tolerance
    elastic_kick_in = 0  # counter to get to limit before elastic kicks in

    def __init__(self, redisConn, myQueue, throttleWindow, throttleLimit,
                 moderate=False, windowName=None, modName=None,
                 elastic=False, elastic_buffer=0):
        '''
        For best performance, all instances of a throttled queue should have
            the same settings
        Limits outbound flow (pop) from any Redis Queue, does not hinder pushes
        This queue is also temporary, which is why is is a bit complex

        @param redis: The redis connection to use
        @param queueClass: The instantiated RedisQueue class
            (Queue, Stack, Priority)
        @param throttleWindow: The time window to throttle pop requests (secs)
        @param throttleLimit: The number of pops allows in a given time window
        @param moderation: Set to True if you would like the queue to have
            a more consistent outbound flow.
        @param windowName: Use a different rolling window key name
        @param modName: Use a different moderate time key name
        @param elastic: When moderated and falling behind, break moderation to
            catch up to desired limit
        @param elastic_buffer: The threshold number for how close we should get
            to the limit when using the elastic catch up
        '''
        self.redis_conn = redisConn
        self.queue = myQueue
        self.window = float(throttleWindow)
        self.limit = float(throttleLimit)

        if windowName is None:
            # default window name
            self.window_key = self.queue.key + self.window_append
        else:
            self.window_key = windowName + self.window_append

        # moderation is useless when only grabbing 1 item in x secs
        if moderate and throttleLimit != 1:
            self.moderation = old_div(self.window, self.limit)
            # used for communicating throttle moderation across queue instances
            if modName is None:
                self.moderate_key = self.queue.key + self.time_append
            else:
                self.moderate_key = modName + self.time_append

            self.elastic = elastic
            self.elastic_buffer = elastic_buffer

    def __len__(self):
        '''
        Return the length of the queue
        '''
        return len(self.queue)

    def clear(self):
        '''
        Clears all data associated with the throttled queue
        '''
        self.redis_conn.delete(self.window_key)
        self.redis_conn.delete(self.moderate_key)
        self.queue.clear()

    def push(self, *args):
        '''
        Push a request into the queue
        '''
        self.queue.push(*args)

    def pop(self, *args):
        '''
        Non-blocking from throttled queue standpoint, tries to return a
        queue pop request, only will return a request if
        the given time window has not been exceeded

        @return: The item if the throttle limit has not been hit,
        otherwise None
        '''
        if self.allowed():
            if self.elastic_kick_in < self.limit:
                self.elastic_kick_in += 1
            return self.queue.pop(*args)
        else:
            return None

    '''
    Original Redis Throttle implementation from
    http://opensourcehacker.com/2014/07/09/rolling-time-window-counters-with-redis-and-mitigating-botnet-driven-login-attacks/
    Modified heavily to fit our class needs, plus locking
    mechanisms around the operations
    '''

    def allowed(self):
        '''
        Check to see if the pop request is allowed

        @return: True means the maximum was not been reached for the current
            time window, thus allowing what ever operation follows
        '''
        # Expire old keys (hits)
        expires = time.time() - self.window
        self.redis_conn.zremrangebyscore(self.window_key, '-inf', expires)

        # check if we are hitting too fast for moderation
        if self.moderation:
            with self.redis_conn.pipeline() as pipe:
                try:
                    pipe.watch(self.moderate_key)  # ---- LOCK
                    # from this point onward if no errors are raised we
                    # successfully incremented the counter

                    curr_time = time.time()
                    if self.is_moderated(curr_time, pipe) and not \
                            self.check_elastic():
                        return False

                    # passed the moderation limit, now check time window
                    # If we have less keys than max, update out moderate key
                    if self.test_hits():
                        # this is a valid transaction, set the new time
                        pipe.multi()
                        pipe.set(name=self.moderate_key,
                                 value=str(curr_time),
                                 ex=int(self.window * 2))
                        pipe.execute()
                        return True

                except WatchError:
                    # watch was changed, another thread just incremented
                    # the value
                    return False

        # If we currently have more keys than max,
        # then limit the action
        else:
            return self.test_hits()

        return False

    def check_elastic(self):
        '''
        Checks if we need to break moderation in order to maintain our desired
        throttle limit

        @return: True if we need to break moderation
        '''
        if self.elastic and self.elastic_kick_in == self.limit:
            value = self.redis_conn.zcard(self.window_key)
            if self.limit - value > self.elastic_buffer:
                return True
        return False

    def is_moderated(self, curr_time, pipe):
        '''
        Tests to see if the moderation limit is not exceeded

        @return: True if the moderation limit is exceeded
        '''
        # get key, otherwise default the moderate key expired and
        # we dont care
        value = pipe.get(self.moderate_key)
        if value is None:
            value = 0.0
        else:
            value = float(value)

        # check moderation difference
        if (curr_time - value) < self.moderation:
            return True

        return False

    def test_hits(self):
        '''
        Tests to see if the number of throttle queue hits is within our limit

        @return: True if the queue was below the limit AND atomically updated
        '''
        with self.redis_conn.pipeline() as pipe:
            try:
                pipe.watch(self.window_key)  # ---- LOCK
                value = self.redis_conn.zcard(self.window_key)
                if value < self.limit:
                    # push value into key
                    now = time.time()
                    pipe.multi()
                    pipe.zadd(self.window_key, {now: now})
                    # expire it if it hasnt been touched in a while
                    pipe.expire(self.window_key, int(self.window * 2))
                    pipe.execute()

                    return True

            except WatchError:
                # watch was changed, another thread just messed with the
                # queue so we can't tell if our result is ok
                pass

        return False



# -*- coding: utf-8 -*-

# Define here the models for your scraped items

from scrapy import Item, Field


class RawResponseItem(Item):
    appid = Field()
    crawlid = Field()
    url = Field()
    responseUrl = Field()
    statusCode = Field()
    success = Field()
    exception = Field()
    encoding = Field()
    playgroundId = Field()
    attrs = Field()


class MenuResponseItem(RawResponseItem):
    groupCategoryName = Field()
    groupName = Field()
    groupUrl = Field()


class ProductResponseItem(RawResponseItem):
    productUrl = Field()
    groupId = Field()
    imageUrl = Field()
    name = Field()
    details = Field()
    price = Field()


class ProductDetailsResponseItem(RawResponseItem):
    productUrl = Field()
    groupId = Field()
    imageUrl = Field()
    name = Field()
    details = Field()



'''
Offline utility tests
'''
from __future__ import print_function
import six
import sys
from os import path
from unittest import TestCase

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

from scutils.settings_wrapper import SettingsWrapper


class TestSettingsWrapper(TestCase):

    defaults = {"STRING": "stuff", "DICT": {"value": "other stuff"}}

    def setUp(self):
        self.wrapper = SettingsWrapper()
        self.wrapper.my_settings = {}

    def test_no_defaults(self):
        self.wrapper.my_settings = {}
        self.wrapper._load_defaults()
        sets = self.wrapper.settings()
        self.assertEqual(sets, {})

    def test_load_default(self):
        self.wrapper._load_defaults("default_settings.py")
        sets = self.wrapper.settings()
        self.assertEqual(sets, self.defaults)

    def test_no_override(self):
        # test no prior defaults
        self.wrapper.my_settings = {}
        self.wrapper._load_custom()
        sets = self.wrapper.settings()
        self.assertEqual(sets, {})

        self.wrapper._load_defaults("default_settings.py")
        self.wrapper._load_custom()
        sets = self.wrapper.settings()
        self.assertEqual(sets, self.defaults)

    def test_override_default(self):
        self.wrapper._load_defaults("default_settings.py")
        self.wrapper._load_custom("override_defaults.py")
        sets = self.wrapper.settings()
        actual = {
            'NEW_DICT': {
                'other': 'stuff'
            },
            'MY_STRING': 'cool',
            'DICT': {
                'append': 'value',
                'value': 'override'
            },
            'STRING': 'my stuff',
            'NEW_LIST': ['item1']
        }
        self.assertEqual(sets, actual)

    def test_load_string(self):
        s = """STRING = \"my stuff\"\nMY_STRING = \"cool\"\nNEW_LIST = [\'item2\']"""

        real = {
            'STRING': 'my stuff',
            'MY_STRING': 'cool',
            'NEW_LIST': ['item2']
        }

        sets = self.wrapper.load_from_string(s)
        six.assertCountEqual(self, real, sets)



import sys
import redis
from bhfutils.scutils.redis_queue import RedisStack, RedisQueue, RedisPriorityQueue
import argparse


def main():
    parser = argparse.ArgumentParser(description='Example Redis Queues.')
    parser.add_argument('-r', '--redis-host', action='store', default='scdev',
                        help="The Redis host ip")
    parser.add_argument('-rp', '--redis-port', action='store', default='6379',
                        help="The Redis port")
    parser.add_argument('-rP', '--redis-password', action='store', default=None,
                        help="The Redis password")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-q', '--queue', action='store_true', help="Use a RedisQueue")
    group.add_argument('-s', '--stack', action='store_true',
                        help="Use a RedisStack")
    group.add_argument('-p', '--priority', action='store_true',
                        help="Use a RedisPriorityQueue")

    args = vars(parser.parse_args())

    host = args['redis_host']
    port = args['redis_port']
    password = args['redis_password']
    redis_conn = redis.Redis(host=host, port=port, password=password, decode_responses=True)

    if args['queue']:
        queue = RedisQueue(redis_conn, "my_key")
    elif args['stack']:
        queue = RedisStack(redis_conn, "my_key")
    elif args['priority']:
        queue = RedisPriorityQueue(redis_conn, "my_key")

    print("Using " + queue.__class__.__name__)

    if isinstance(queue, RedisPriorityQueue):
        queue.push("item1", 50)
        queue.push("item2", 100)
        queue.push("item3", 20)
    else:
        queue.push("item1")
        queue.push("item2")
        queue.push("item3")

    print("Pop 1 " + queue.pop())
    print("Pop 2 " + queue.pop())
    print("Pop 3 " + queue.pop())


if __name__ == "__main__":
    sys.exit(main())



import sys
import time
import redis
import socket
from os import path
from redis.exceptions import ConnectionError

from bhfutils.scutils.log_factory import LogFactory
from bhfutils.scutils.stats_collector import StatsCollector


class RedisStatsMiddleware(object):

    def __init__(self, settings):
        self.setup(settings)

    def setup(self, settings):
        '''
        Does the actual setup of the middleware
        '''
        # set up the default sc logger
        my_level = settings.get('SC_LOG_LEVEL', 'INFO')
        my_name = settings.get('SClogger_NAME', 'sc-logger')
        my_output = settings.get('SC_LOG_STDOUT', True)
        my_json = settings.get('SC_LOG_JSON', False)
        my_dir = settings.get('SC_LOG_DIR', 'logs')
        my_bytes = settings.get('SC_LOG_MAX_BYTES', '10MB')
        my_file = settings.get('SC_LOG_FILE', 'main.log')
        my_backups = settings.get('SC_LOG_BACKUPS', 5)

        self.logger = LogFactory.get_instance(json=my_json,
                                              name=my_name,
                                              stdout=my_output,
                                              level=my_level,
                                              dir=my_dir,
                                              file=my_file,
                                              bytes=my_bytes,
                                              backups=my_backups)

        self.settings = settings
        self.stats_dict = {}

        # set up redis
        self.redis_conn = redis.Redis(host=settings.get('REDIS_HOST'),
                                      port=settings.get('REDIS_PORT'),
                                      db=settings.get('REDIS_DB'),
                                      password=settings.get('REDIS_PASSWORD'),
                                      decode_responses=True,
                                      socket_timeout=settings.get('REDIS_SOCKET_TIMEOUT'),
                                      socket_connect_timeout=settings.get('REDIS_SOCKET_TIMEOUT'))

        try:
            self.redis_conn.info()
            self.logger.debug("Connected to Redis in ScraperHandler")
        except ConnectionError:
            self.logger.error("Failed to connect to Redis in Stats Middleware")
            # plugin is essential to functionality
            sys.exit(1)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def _get_hostname(self):
        '''
        Gets the hostname of the machine the spider is running on

        @return: the hostname of the machine
        '''
        return socket.gethostname()

    def _setup_stats_status_codes(self, spider_name):
        '''
        Sets up the status code stats collectors
        '''
        self.stats_dict[spider_name] = {
            'status_codes': {}
        }
        self.stats_dict[spider_name]['status_codes'] = {}
        hostname = self._get_hostname()
        # we chose to handle 504's here as well as in the middleware
        # in case the middleware is disabled
        for status_code in self.settings['STATS_RESPONSE_CODES']:
            temp_key = 'stats:crawler:{h}:{n}:{s}'.format(
                h=hostname, n=spider_name, s=status_code)
            self.stats_dict[spider_name]['status_codes'][status_code] = {}
            for item in self.settings['STATS_TIMES']:
                try:
                    time = getattr(StatsCollector, item)

                    self.stats_dict[spider_name]['status_codes'][status_code][time] = StatsCollector \
                        .get_rolling_time_window(
                        redis_conn=self.redis_conn,
                        key='{k}:{t}'.format(k=temp_key, t=time),
                        window=time,
                        cycle_time=self.settings['STATS_CYCLE'])
                    self.logger.debug("Set up status code {s}, {n} spider," \
                                      " host {h} Stats Collector '{i}'" \
                                      .format(h=hostname, n=spider_name, s=status_code, i=item))
                except AttributeError as e:
                    self.logger.warning("Unable to find Stats Time '{s}'" \
                                        .format(s=item))
            total = StatsCollector.get_hll_counter(redis_conn=self.redis_conn,
                                                   key='{k}:lifetime'.format(k=temp_key),
                                                   cycle_time=self.settings['STATS_CYCLE'],
                                                   roll=False)
            self.logger.debug("Set up status code {s}, {n} spider," \
                              "host {h} Stats Collector 'lifetime'" \
                              .format(h=hostname, n=spider_name, s=status_code))
            self.stats_dict[spider_name]['status_codes'][status_code]['lifetime'] = total

    def process_spider_input(self, response, spider):
        '''
        Ensures the meta data from the response is passed
        through in any Request's generated from the spider
        '''
        self.logger.debug("processing redis stats middleware")
        if self.settings['STATS_STATUS_CODES']:
            if spider.name not in self.stats_dict:
                self._setup_stats_status_codes(spider.name)

            if 'status_codes' in self.stats_dict[spider.name]:
                code = response.status
                if code in self.settings['STATS_RESPONSE_CODES']:
                    for key in self.stats_dict[spider.name]['status_codes'][code]:
                        try:
                            if key == 'lifetime':
                                unique = response.url + str(response.status) \
                                         + str(time.time())
                                self.stats_dict[spider.name]['status_codes'][code][key].increment(unique)
                            else:
                                self.stats_dict[spider.name]['status_codes'][code][key].increment()
                        except Exception as e:
                            self.logger.warn("Error in spider redis stats")
                    self.logger.debug("Incremented status_code '{c}' stats" \
                                      .format(c=code))



'''
Offline tests
'''
import mock
from builtins import object
from unittest import TestCase
from mock import MagicMock
from scrapy.http import Request
from scrapy.utils.reqser import request_to_dict
from crawling.distributed_scheduler import DistributedScheduler
from bhfutils.scutils import RedisThrottledQueue


class ThrottleMixin(object):

    @mock.patch('crawling.distributed_scheduler.DistributedScheduler' \
                '.update_ipaddress')
    @mock.patch('crawling.distributed_scheduler.DistributedScheduler' \
                '.setup_zookeeper')
    def setUp(self, u, z):
        self.scheduler = DistributedScheduler(MagicMock(), False, 60, 10, 3,
                                              MagicMock(), 10, 60, False, 60,
                                              False, False, '.*', True, 3600,
                                              None, 600, 600)
        self.scheduler.open(MagicMock())
        self.scheduler.my_ip = 'ip'
        self.scheduler.spider.name = 'link'
        self.req = None

    def get_request(self):
        req = None

        # required
        req = Request('http://ex.com')
        req.meta['crawlid'] = "abc123"
        req.meta['appid'] = "myapp"

        req.meta['url'] = "http://ex.com"
        req.meta['spiderid'] = "link"
        req.meta["attrs"] = None
        req.meta["allowed_domains"] = None
        req.meta["allow_regex"] = None
        req.meta["deny_regex"] = None
        req.meta["deny_extensions"] = None
        req.meta['curdepth'] = 0
        req.meta["maxdepth"] = 0
        req.meta["domain_max_pages"] = None
        req.meta['priority'] = 0
        req.meta['retry_times'] = 0
        req.meta['expires'] = 0
        req.meta['useragent'] = None
        req.meta['cookie'] = None

        return req


class TestDistributedSchedulerEnqueueRequest(ThrottleMixin, TestCase):

    @mock.patch('time.time', return_value=5)
    def test_enqueue_request(self, t):
        self.req = self.get_request()

        # test request already seen
        self.scheduler.dupefilter.request_seen = MagicMock(return_value=True)
        self.assertEqual(self.scheduler.enqueue_request(self.req), None)

        # test global page limit reached
        self.scheduler.global_page_per_domain_filter = MagicMock(return_value=True)
        self.assertEqual(self.scheduler.enqueue_request(self.req), None)

        # test per domain page limit reached
        self.scheduler.domain_max_page_filter = MagicMock(return_value=True)
        self.assertEqual(self.scheduler.enqueue_request(self.req), None)

        # test request not expiring and queue seen
        self.scheduler.queue_keys = ['link:ex.com:queue']
        self.extract = MagicMock(return_value={"domain": 'ex', "suffix": 'com'})
        self.scheduler.is_blacklisted = MagicMock(return_value=False)
        self.scheduler.dupefilter.request_seen = MagicMock(return_value=False)
        self.scheduler.global_page_per_domain_filter.request_page_limit_reached = MagicMock(return_value=False)
        self.scheduler.domain_max_page_filter.request_page_limit_reached = MagicMock(return_value=False)
        self.scheduler.queue_dict['link:ex.com:queue'] = [MagicMock(), 0]
        self.scheduler.queue_dict['link:ex.com:queue'][0].push = MagicMock(
                                                    side_effect=Exception("1"))
        try:
            self.scheduler.enqueue_request(self.req)
            # this should not be reached
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "1")

        # test request not expiring and queue not seen
        self.scheduler.redis_conn.zadd = MagicMock(side_effect=Exception("2"))
        self.scheduler.queue_keys = []
        try:
            self.scheduler.enqueue_request(self.req)
            # this should not be reached
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "2")

        # test whole domain blacklisted, but we allow it
        self.scheduler.black_domains = ['ex.com']
        try:
            self.scheduler.enqueue_request(self.req)
            # this should not be reached
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "2")

        # test dont allow blacklist domains back into the queue
        self.scheduler.backlog_blacklist = False
        self.scheduler.enqueue_request(self.req)

        # test allow domain back into queue since not blacklisted
        self.scheduler.black_domains = ['ex2.com']
        self.scheduler.backlog_blacklist = False
        try:
            self.scheduler.enqueue_request(self.req)
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "2")
        # reset
        self.scheduler.black_domains = []
        self.scheduler.backlog_blacklist = True

        # test request expired
        self.req.meta['expires'] = 1
        self.assertEqual(self.scheduler.enqueue_request(self.req), None)

        # test request blacklisted via stop or expire from redis-monitor
        self.scheduler.is_blacklisted = MagicMock(return_value=True)
        self.assertEqual(self.scheduler.enqueue_request(self.req), None)


class TestDistributedSchedulerFindItem(ThrottleMixin, TestCase):

    def test_find_item(self):
        # test finding an item
        self.scheduler.queue_keys = ["link:ex.com:queue"]
        self.scheduler.queue_dict = {"link:ex.com:queue": [MagicMock(), 0]}
        self.scheduler.queue_dict["link:ex.com:queue"][0].pop = MagicMock(return_value='item')
        self.assertEqual(self.scheduler.find_item(), 'item')

        # test failed to find an item
        self.scheduler.queue_dict["link:ex.com:queue"][0].pop = MagicMock(return_value=None)
        self.assertEqual(self.scheduler.find_item(), None)

        # test skip due to blacklist
        self.scheduler.black_domains = ['ex.com']
        self.scheduler.queue_dict["link:ex.com:queue"][0].pop = MagicMock(side_effect=Exception("bad"))
        self.assertEqual(self.scheduler.find_item(), None) # should also not raise exception


class TestDistributedSchedulerRequestFromFeed(ThrottleMixin, TestCase):
    def test_request_from_feed(self):
        self.req = self.get_request()
        feed = {
            "url": "http://ex.com",
            "crawlid": "abc123",
            "appid": "myapp",
            "spiderid": "link",
        }
        out = self.scheduler.request_from_feed(feed)
        self.assertEqual(out.url, 'http://ex.com')
        for key in out.meta:
            self.assertEqual(out.meta[key], self.req.meta[key])


class TestDistributedSchedulerNextRequest(ThrottleMixin, TestCase):

    @mock.patch('time.time', return_value=5)
    def test_next_request(self, t):
        self.req = self.get_request()

        # test update queues
        self.scheduler.update_time = 1
        self.scheduler.update_interval = 2
        self.scheduler.create_queues = MagicMock(side_effect=Exception("q"))
        try:
            self.scheduler.next_request()
            # this should not be reached
            self.assertFalse(True)
        except Exception as e:
            print("str", e, e == "q", e == "'q'", e == 'q')
            self.assertEqual(str(e), "q")

        # test update ip address
        self.scheduler.update_time = 4
        self.scheduler.update_ipaddress = MagicMock(side_effect=Exception("ip"))
        self.scheduler.update_ip_time = 1
        self.scheduler.ip_update_interval = 2
        try:
            self.scheduler.next_request()
            # this should not be reached
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "ip")

        # test request from feed
        feed = {
            "url": "http://ex.com",
            "crawlid": "abc123",
            "appid": "myapp",
            "spiderid": "link",
        }
        self.scheduler.find_item = MagicMock(return_value=feed)
        out = self.scheduler.next_request()
        self.assertEqual(out.url, 'http://ex.com')
        for key in out.meta:
            self.assertEqual(out.meta[key], self.req.meta[key])

        # test request from feed with cookies
        feed = {
            "url": "http://ex.com",
            "crawlid": "abc123",
            "appid": "myapp",
            "spiderid": "link",
            "cookie": "authenticated=true;privacy=10"
        }
        self.req.meta['cookie'] = "authenticated=true;privacy=10"  # add cookie to req since we are not testing this
        self.scheduler.find_item = MagicMock(return_value=feed)
        out = self.scheduler.next_request()
        self.assertEqual(out.url, 'http://ex.com')
        for key in out.meta:
            self.assertEqual(out.meta[key], self.req.meta[key])
        self.assertEqual(out.cookies, self.scheduler.parse_cookie(feed["cookie"]))
        self.req.meta['cookie'] = None  # reset

        # test request from serialized request
        exist_req = Request('http://ex.com')
        exist_req.meta["crawlid"] = "abc123"
        exist_req.meta["appid"] = "myapp"
        exist_req.meta["spiderid"] = "link"
        exist_item = request_to_dict(exist_req)
        self.scheduler.find_item = MagicMock(return_value=exist_item)
        out = self.scheduler.next_request()
        self.assertEqual(out.url, 'http://ex.com')
        for key in out.meta:
            self.assertEqual(out.meta[key], exist_req.meta[key])

        # test request from serialized request with supplied cookie
        exist_req = Request('http://ex.com', cookies={'auth':'101'})
        exist_item = request_to_dict(exist_req)
        self.scheduler.find_item = MagicMock(return_value=exist_item)
        out = self.scheduler.next_request()
        self.assertEqual(out.url, 'http://ex.com')
        for key in out.meta:
            self.assertEqual(out.meta[key], exist_req.meta[key])
        self.assertEqual(out.cookies, exist_req.cookies)
        self.req.meta['cookie'] = None  # reset

        # test request from serialized request with meta cookie
        exist_req = Request('http://ex.com')
        exist_req.meta["crawlid"] = "abc123"
        exist_req.meta["appid"] = "myapp"
        exist_req.meta["spiderid"] = "link"
        exist_req.meta["cookie"] = {'authenticated': False, 'privacy':9}
        exist_item = request_to_dict(exist_req)
        self.scheduler.find_item = MagicMock(return_value=exist_item)
        out = self.scheduler.next_request()
        self.assertEqual(out.url, 'http://ex.com')
        for key in out.meta:
            self.assertEqual(out.meta[key], exist_req.meta[key])
        self.assertEqual(out.cookies, exist_req.meta['cookie'])

        # test didn't get item
        self.scheduler.find_item = MagicMock(return_value=None)
        self.assertEqual(self.scheduler.next_request(), None)


class TestDistributedSchedulerChangeConfig(ThrottleMixin, TestCase):

    def test_change_config(self):
        good_string = ""\
          "domains:\n"\
          "  dmoztools.net:\n"\
          "      window: 60\n"\
          "      hits: 60\n"\
          "      scale: 1.0\n"\
          "  wikipedia.org:\n"\
          "      window: 60\n"\
          "      hits: 30"
        bad_string1 = "blahrg\ndumb"
        bad_string2 = None
        bad_string3 = ""

        self.scheduler.load_domain_config = MagicMock(side_effect=Exception("1"))
        self.scheduler.error_config = MagicMock(side_effect=Exception("2"))
        self.scheduler.create_queues = MagicMock(side_effect=Exception("3"))

        try:
            self.scheduler.change_config(good_string)
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "1")

        try:
            self.scheduler.change_config(bad_string1)
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "1")

        try:
            self.scheduler.change_config(bad_string2)
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "2")

        try:
            self.scheduler.change_config(bad_string3)
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "2")


class TestDistributedSchedulerLoadDomainConfig(ThrottleMixin, TestCase):

    def test_load_domain_config(self):
        good_yaml_dict = {
            "blacklist": ["blah.com"],
            "domains": {
                "ex1.com": {
                    "window": 60,
                    "hits": 10,
                    "scale": 1
                },
                "ex2.com": {
                    "window": 60,
                    "hits": 10
                }
            }
        }
        bad_yaml_dict1 = {
            "blarg": ['stuff']
        }
        bad_yaml_dict2 = {
            "domains": {
                "stuff.com": {
                    "window": 60
                },
                "stuff2.com": {
                    "hits": 10
                }
            }
        }
        result_yaml = {}
        result_yaml['ex1.com'] = good_yaml_dict['domains']['ex1.com']
        result_yaml['ex2.com'] = good_yaml_dict['domains']['ex2.com']

        # correctly loaded
        self.scheduler.load_domain_config(good_yaml_dict)
        self.assertEqual(self.scheduler.domain_config, result_yaml)
        self.assertEqual(self.scheduler.black_domains, ["blah.com"])

        # both are not correct yaml setups
        self.scheduler.load_domain_config(bad_yaml_dict1)
        self.assertEqual(self.scheduler.domain_config, {})

        self.scheduler.load_domain_config(bad_yaml_dict2)
        self.assertEqual(self.scheduler.domain_config, {})


class TestDistributedSchedulerUpdateDomainQueues(ThrottleMixin, TestCase):

    def test_update_domain_queues(self):
        # test without scale factor
        self.scheduler.domain_config = {
            "ex1.com": {
                "window": 50,
                "hits": 10,
                "scale": 1
            }
        }
        q = RedisThrottledQueue(MagicMock(), MagicMock(), 100, 100)
        self.scheduler.queue_dict = {'link:ex1.com:queue': [q, 0]}

        self.scheduler.update_domain_queues()
        self.assertEqual(self.scheduler.queue_dict['link:ex1.com:queue'][0].window, 50)
        self.assertEqual(self.scheduler.queue_dict['link:ex1.com:queue'][0].limit, 10)

        # test with scale factor
        self.scheduler.domain_config = {
            "ex2.com": {
                "window": 50,
                "hits": 10,
                "scale": 0.5
            }
        }
        q = RedisThrottledQueue(MagicMock(), MagicMock(), 100, 100)
        self.scheduler.queue_dict = {'link:ex2.com:queue': [q, 0]}

        self.scheduler.update_domain_queues()
        self.assertEqual(self.scheduler.queue_dict['link:ex2.com:queue'][0].window, 50)
        # the scale factor effects the limit only
        self.assertEqual(self.scheduler.queue_dict['link:ex2.com:queue'][0].limit, 5)


class TestDistributedSchedulerErrorConfig(ThrottleMixin, TestCase):

    def test_error_config(self):
        self.scheduler.domain_config = {
            "ex1.com": {
                "window": 50,
                "hits": 10
            }
        }
        self.scheduler.window = 7
        self.scheduler.hits = 5
        q = RedisThrottledQueue(MagicMock(), MagicMock(), 100, 100)
        self.scheduler.queue_dict = {'link:ex1.com:queue': [q, 0]}

        self.scheduler.error_config('stuff')

        self.assertEqual(self.scheduler.queue_dict['link:ex1.com:queue'][0].window, 7)
        self.assertEqual(self.scheduler.queue_dict['link:ex1.com:queue'][0].limit, 5)
        self.assertEqual(self.scheduler.domain_config, {})


class TestDistributedSchedulerFitScale(ThrottleMixin, TestCase):

    def test_fit_scale(self):
        # assert max
        self.assertEqual(self.scheduler.fit_scale(1.1), 1.0)

        # assert min
        self.assertEqual(self.scheduler.fit_scale(-1.9), 0.0)

        # assert normal
        self.assertEqual(self.scheduler.fit_scale(0.51), 0.51)


class TestDistributedSchedulerCreateQueues(ThrottleMixin, TestCase):

    def test_create_queues(self):
        queues = ['link:ex1:queue', 'link:ex2:queue', 'link:ex3:queue']
        self.scheduler.redis_conn.keys = MagicMock(return_value=queues)

        # test basic
        self.scheduler.create_queues()
        expected = ['ex1:throttle_window',
                    'ex2:throttle_window',
                    'ex3:throttle_window']
        for key in self.scheduler.queue_dict:
            self.assertTrue(self.scheduler.queue_dict[key][0].window_key
                            in expected)

        # test type
        self.scheduler.add_type = True
        self.scheduler.queue_dict = {}
        expected = ['link:ex1:throttle_window',
                    'link:ex2:throttle_window',
                    'link:ex3:throttle_window']
        self.scheduler.create_queues()
        for key in self.scheduler.queue_dict:
            self.assertTrue(self.scheduler.queue_dict[key][0].window_key
                            in expected)

        # test ip
        self.scheduler.add_ip = True
        self.scheduler.add_type = False
        self.scheduler.queue_dict = {}
        expected = ['ip:ex1:throttle_window',
                    'ip:ex2:throttle_window',
                    'ip:ex3:throttle_window']
        self.scheduler.create_queues()
        for key in self.scheduler.queue_dict:
            self.assertTrue(self.scheduler.queue_dict[key][0].window_key
                            in expected)

        # test type and ip
        self.scheduler.add_ip = True
        self.scheduler.add_type = True
        self.scheduler.queue_dict = {}
        expected = ['link:ip:ex1:throttle_window',
                    'link:ip:ex2:throttle_window',
                    'link:ip:ex3:throttle_window']
        self.scheduler.create_queues()
        for key in self.scheduler.queue_dict:
            self.assertTrue(self.scheduler.queue_dict[key][0].window_key
                            in expected)


class TestDistributedSchedulerExpireQueues(ThrottleMixin, TestCase):

    @mock.patch('time.time', return_value=5)
    def test_expire_queues(self, t):
        self.scheduler.queue_dict = {'key1':[MagicMock(), 3]}
        self.scheduler.queue_keys = []

        # assert not expired
        self.scheduler.queue_timeout = 3
        self.scheduler.expire_queues()
        self.assertTrue('key1' in self.scheduler.queue_dict)

        # assert expired
        self.scheduler.queue_timeout = 1
        self.scheduler.expire_queues()
        self.assertTrue('key1' not in self.scheduler.queue_dict.keys())

        # assert remove from queue_keys also
        self.scheduler.queue_dict = {'key1':[MagicMock(), 3]}
        self.scheduler.queue_keys = ['key1']
        self.scheduler.expire_queues()
        self.assertTrue('key1' not in self.scheduler.queue_dict.keys())
        self.assertTrue('key1' not in self.scheduler.queue_keys)


class TestDistributedSchedulerParseCookie(ThrottleMixin, TestCase):

    def test_parse_cookie(self):
        cookie = "blah=stuff; expires=Thu, 18 May 2017 12:42:29 GMT; path"\
        "=/; Domain=.domain.com"
        result = {
            "blah": "stuff",
            "expires": "Thu, 18 May 2017 12:42:29 GMT",
            "path": "/",
            "Domain": ".domain.com",
        }

        self.assertEqual(result, self.scheduler.parse_cookie(cookie))



import sys
import time
import redis
import socket
from os import path
from builtins import str
from builtins import object

from twisted.web._newclient import ResponseFailed
from twisted.internet import defer
from twisted.internet.error import TimeoutError, DNSLookupError, \
        ConnectionRefusedError, ConnectionDone, ConnectError, \
        ConnectionLost, TCPTimedOutError
from redis.exceptions import ConnectionError

from bhfutils.scutils.log_factory import LogFactory
from bhfutils.scutils.stats_collector import StatsCollector


class LogRetryMiddleware(object):

    EXCEPTIONS_TO_RETRY = (defer.TimeoutError, TimeoutError, DNSLookupError,
                           ConnectionRefusedError, ConnectionDone, ConnectError,
                           ConnectionLost, TCPTimedOutError, ResponseFailed,
                           IOError)

    def __init__(self, settings):
        self.setup(settings)

    def setup(self, settings):
        '''
        Does the actual setup of the middleware
        '''
        # set up the default sc logger
        my_level = settings.get('SC_LOG_LEVEL', 'INFO')
        my_name = settings.get('SC_LOGGER_NAME', 'sc-logger')
        my_output = settings.get('SC_LOG_STDOUT', True)
        my_json = settings.get('SC_LOG_JSON', False)
        my_dir = settings.get('SC_LOG_DIR', 'logs')
        my_bytes = settings.get('SC_LOG_MAX_BYTES', '10MB')
        my_file = settings.get('SC_LOG_FILE', 'main.log')
        my_backups = settings.get('SC_LOG_BACKUPS', 5)

        self.logger = LogFactory.get_instance(json=my_json,
                                         name=my_name,
                                         stdout=my_output,
                                         level=my_level,
                                         dir=my_dir,
                                         file=my_file,
                                         bytes=my_bytes,
                                         backups=my_backups)

        self.retry_http_codes = set(int(x) for x in
                                    settings.getlist('RETRY_HTTP_CODES'))

        # stats setup
        self.stats_dict = {}
        self.settings = settings
        self.name = self.settings['SPIDER_NAME']
        if self.settings['STATS_STATUS_CODES']:
            self.redis_conn = redis.Redis(host=self.settings.get('REDIS_HOST'),
                                          port=self.settings.get('REDIS_PORT'),
                                          db=settings.get('REDIS_DB'),
                                          password=self.settings.get('REDIS_PASSWORD'),
                                          decode_responses=True,
                                          socket_timeout=self.settings.get('REDIS_SOCKET_TIMEOUT'),
                                          socket_connect_timeout=self.settings.get('REDIS_SOCKET_TIMEOUT'))

            try:
                self.redis_conn.info()
                self.logger.debug("Connected to Redis in LogRetryMiddleware")
            except ConnectionError:
                self.logger.error("Failed to connect to Redis in LogRetryMiddleware")
                # plugin is essential to functionality
                sys.exit(1)

            self._setup_stats_status_codes()

    @classmethod
    def from_settings(cls, settings):
        return cls(settings)

    @classmethod
    def from_crawler(cls, crawler):
        # hack to update passed in settings
        crawler.settings.frozen = False
        crawler.settings.set('SPIDER_NAME', crawler.spider.name)
        return cls.from_settings(crawler.settings)

    def process_exception(self, request, exception, spider):
        if isinstance(exception, self.EXCEPTIONS_TO_RETRY):
            self._log_retry(request, exception, spider)
            self._increment_504_stat(request)

    def _log_retry(self, request, exception, spider):
        extras = {}
        extras['logger'] = self.logger.name
        extras['error_request'] = request
        extras['error_reason'] = exception
        extras['retry_count'] = request.meta.get('retry_times', 0)
        extras['status_code'] = 504
        extras['appid'] = request.meta['appid']
        extras['crawlid'] = request.meta['crawlid']
        extras['url'] = request.url

        self.logger.error('Scraper Retry', extra=extras)

    def _setup_stats_status_codes(self):
        '''
        Sets up the status code stats collectors
        '''
        hostname = self._get_hostname()
        # we chose to handle 504's here as well as in the middleware
        # in case the middleware is disabled
        self.logger.debug("Setting up log retry middleware stats")
        status_code = 504
        temp_key = 'stats:crawler:{h}:{n}:{s}'.format(
            h=hostname, n=self.name, s=status_code)
        for item in self.settings['STATS_TIMES']:
            try:
                time = getattr(StatsCollector, item)

                self.stats_dict[time] = StatsCollector \
                        .get_rolling_time_window(
                                redis_conn=self.redis_conn,
                                key='{k}:{t}'.format(k=temp_key, t=time),
                                window=time,
                                cycle_time=self.settings['STATS_CYCLE'])
                self.logger.debug("Set up LRM status code {s}, {n} spider,"\
                    " host {h} Stats Collector '{i}'"\
                    .format(h=hostname, n=self.name, s=status_code, i=item))
            except AttributeError as e:
                self.logger.warning("Unable to find Stats Time '{s}'"\
                        .format(s=item))
        total = StatsCollector.get_hll_counter(redis_conn=self.redis_conn,
                        key='{k}:lifetime'.format(k=temp_key),
                        cycle_time=self.settings['STATS_CYCLE'],
                        roll=False)
        self.logger.debug("Set up status code {s}, {n} spider,"\
                    "host {h} Stats Collector 'lifetime'"\
                        .format(h=hostname, n=self.name, s=status_code))
        self.stats_dict['lifetime'] = total

    def _get_hostname(self):
        '''
        Gets the hostname of the machine the spider is running on

        @return: the hostname of the machine
        '''
        return socket.gethostname()

    def _increment_504_stat(self, request):
        '''
        Increments the 504 stat counters

        @param request: The scrapy request in the spider
        '''
        for key in self.stats_dict:
            if key == 'lifetime':
                unique = request.url + str(time.time())
                self.stats_dict[key].increment(unique)
            else:
                self.stats_dict[key].increment()
        self.logger.debug("Incremented status_code '504' stats")



from __future__ import print_function
from future import standard_library

import sys
import logging
import argparse
from time import sleep
from builtins import object
from threading import Thread

standard_library.install_aliases()
from kazoo.client import KazooClient, KazooState
from kazoo.exceptions import ZookeeperError, KazooException

log = logging.getLogger(__name__)


class ZookeeperWatcher(object):
    zoo_client = None  # The KazooClient to manage the config
    point_path = None  # Zookeeper path to pointed to file
    pointed_at_expired = None  # is True when the assignment has been set to
    # None but we cannot remove the config listener
    valid_handler = None  # the function to call when the validity changes
    config_handler = None  # the function to call when the config changes
    error_handler = None  # the function to call when an error occurs in reading
    valid_file = False  # the current state of the ConfigWatcher with ZK
    do_not_restart = False  # used when closing via ^C
    old_data = ''  # The current file contents, to see if a change occurred
    old_pointed = ''  # the current pointed path, to see if change occurred

    INVALID_PATH = "Invalid pointer path"
    INVALID_GET = "Invalid get on file path"
    BAD_CONNECTION = "Connection interrupted with Zookeeper, re-establishing"

    def __init__(self, hosts, filepath, valid_handler=None,
                 config_handler=None, error_handler=None, pointer=False,
                 ensure=False, valid_init=True):
        '''
        Zookeeper file watcher, used to tell a program their zookeeper file has
        changed. Can be used to watch a single file, or both a file and path of
        its contents. Manages all connections, drops, reconnections for you.

        @param hosts: The zookeeper hosts to use
        @param filepath: The full path to the file to watch
        @param valid_handler: The method to call for a 'is valid' state change
        @param config_handler: The method to call when a content change occurs
        @param error_handler: The method to call when an error occurs
        @param pointer: Set to true if the file contents are actually a path to
                        another zookeeper file, where the real config resides
        @param ensure: Set to true for the ZooWatcher to create the watched file
        @param valid_init: Ensure the client can connect to Zookeeper first try

        Ex 1. /stuff/A: "stuff I care about"
        Ex 2. /stuff/A: "/other/stuff", /other/stuff: "contents I care about"
            - in Ex 2 you care about /other/stuff contents
              but are only aware of your assignment /stuff/A

        You can use this class as any combination of event driven or polling.
        Polling:
            In the main loop of your program, check if is_valid() is
            True, otherwise clear your contents as there is some ZK error.
        Event:
            You will be notified via the various handlers when content changes.
        '''
        self.hosts = hosts
        self.my_file = filepath
        self.pointer = pointer
        self.ensure = ensure
        self.valid_handler = valid_handler
        self.config_handler = config_handler
        self.error_handler = error_handler

        if valid_init:
            # this will throw an exception if it can't start right away
            self.zoo_client = KazooClient(hosts=self.hosts)
            self.zoo_client.start()

        self.threaded_start(no_init=True)

    def threaded_start(self, no_init=False):
        '''
        Spawns a worker thread to set up the zookeeper connection
        '''
        thread = Thread(target=self.init_connections, kwargs={
            'no_init': no_init})
        thread.setDaemon(True)
        thread.start()
        thread.join()

    def init_connections(self, no_init=False):
        '''
        Sets up the initial Kazoo Client and watches
        '''
        success = False
        self.set_valid(False)

        if not no_init:
            if self.zoo_client:
                self.zoo_client.remove_listener(self.state_listener)
                self.old_data = ''
                self.old_pointed = ''

            while not success:
                try:
                    if self.zoo_client is None:
                        self.zoo_client = KazooClient(hosts=self.hosts)
                        self.zoo_client.start()
                    else:
                        # self.zoo_client.stop()
                        self.zoo_client._connection.connection_stopped.set()
                        self.zoo_client.close()
                        self.zoo_client = KazooClient(hosts=self.hosts)
                        self.zoo_client.start()
                except Exception as e:
                    log.error("ZKWatcher Exception: " + e.message)
                    sleep(1)
                    continue

                self.setup()
                success = self.update_file(self.my_file)
                sleep(5)
        else:
            self.setup()
            self.update_file(self.my_file)

    def setup(self):
        '''
        Ensures the path to the watched file exists and we have a state
        listener
        '''
        self.zoo_client.add_listener(self.state_listener)

        if self.ensure:
            self.zoo_client.ensure_path(self.my_file)

    def state_listener(self, state):
        '''
        Restarts the session if we get anything besides CONNECTED
        '''
        if state == KazooState.SUSPENDED:
            self.set_valid(False)
            self.call_error(self.BAD_CONNECTION)
        elif state == KazooState.LOST and not self.do_not_restart:
            self.threaded_start()
        elif state == KazooState.CONNECTED:
            # This is going to throw a SUSPENDED kazoo error
            # which will cause the sessions to be wiped and re established.
            # Used b/c of massive connection pool issues
            self.zoo_client.stop()

    def is_valid(self):
        '''
        @return: True if the currently watch file is valid
        '''
        return self.valid_file

    def ping(self):
        '''
        Simple command to test if the zookeeper session is able to connect
        at this very moment
        '''
        try:
            # dummy ping to ensure we are still connected
            self.zoo_client.server_version()
            return True
        except KazooException as e:
            print(f"error: {e}")
            return False

    def close(self, kill_restart=True):
        '''
        Use when you would like to close everything down
        @param kill_restart= Prevent kazoo restarting from occurring
        '''
        self.do_not_restart = kill_restart
        self.zoo_client.stop()
        self.zoo_client.close()

    def get_file_contents(self, pointer=False):
        '''
        Gets any file contents you care about. Defaults to the main file
        @param pointer: The the contents of the file pointer, not the pointed
        at file
        @return: A string of the contents
        '''
        if self.pointer:
            if pointer:
                return self.old_pointed
            else:
                return self.old_data
        else:
            return self.old_data

    def watch_file(self, event):
        '''
        Fired when changes made to the file
        '''
        if not self.update_file(self.my_file):
            self.threaded_start()

    def update_file(self, path):
        '''
        Updates the file watcher and calls the appropriate method for results
        @return: False if we need to keep trying the connection
        '''
        try:
            # grab the file
            result, stat = self.zoo_client.get(path, watch=self.watch_file)
            result = result.decode('utf-8')
        except ZookeeperError:
            self.set_valid(False)
            self.call_error(self.INVALID_GET)
            return False

        if self.pointer:
            if result is not None and len(result) > 0:
                self.pointed_at_expired = False
                # file is a pointer, go update and watch other file
                self.point_path = result
                if self.compare_pointer(result):
                    self.update_pointed()
            else:
                self.pointed_at_expired = True
                self.old_pointed = ''
                self.old_data = ''
                self.set_valid(False)
                self.call_error(self.INVALID_PATH)
        else:
            # file is not a pointer, return contents
            if self.compare_data(result):
                self.call_config(result)
            self.set_valid(True)

        return True

    def watch_pointed(self, event):
        '''
        Fired when changes made to pointed file
        '''
        self.update_pointed()

    def update_pointed(self):
        '''
        Grabs the latest file contents based on the pointer uri
        '''
        # only grab file if our pointer is still good (not None)
        if not self.pointed_at_expired:
            try:
                conf_string, stat2 = self.zoo_client.get(self.point_path,
                                                         watch=self.watch_pointed)
                conf_string = conf_string.decode('utf-8')
            except ZookeeperError:
                self.old_data = ''
                self.set_valid(False)
                self.pointed_at_expired = True
                self.call_error(self.INVALID_PATH)
                return

            if self.compare_data(conf_string):
                self.call_config(conf_string)
            self.set_valid(True)

    def set_valid(self, boolean):
        '''
        Sets the state and calls the change if needed
        @param bool: The state (true or false)
        '''
        old_state = self.is_valid()
        self.valid_file = boolean

        if old_state != self.valid_file:
            self.call_valid(self.valid_file)

    def call_valid(self, state):
        '''
        Calls the valid change function passed in
        @param valid_state: The new config
        '''
        if self.valid_handler is not None:
            self.valid_handler(self.is_valid())

    def call_config(self, new_config):
        '''
        Calls the config function passed in
        @param new_config: The new config
        '''
        if self.config_handler is not None:
            self.config_handler(new_config)

    def call_error(self, message):
        '''
        Calls the error function passed in
        @param message: The message to throw
        '''
        if self.error_handler is not None:
            self.error_handler(message)

    def compare_data(self, data):
        '''
        Compares the string data
        @return: True if the data is different
        '''
        if self.old_data != data:
            self.old_data = data
            return True
        return False

    def compare_pointer(self, data):
        '''
        Compares the string data
        @return: True if the data is different
        '''
        if self.old_pointed != data:
            self.old_pointed = data
            return True
        return False


def main():
    parser = argparse.ArgumentParser(
        description="Zookeeper file watcher")
    parser.add_argument('-z', '--zoo-keeper', action='store', required=True,
                        help="The Zookeeper connection <host>:<port>")
    parser.add_argument('-f', '--file', action='store', required=True,
                        help="The full path to the file to watch in Zookeeper")
    parser.add_argument('-p', '--pointer', action='store_const', const=True,
                        help="The file contents point to another file")
    parser.add_argument('-s', '--sleep', nargs='?', const=1, default=1,
                        type=int, help="The time to sleep between poll checks")
    parser.add_argument('-v', '--valid-init', action='store_false',
                        help="Do not ensure zookeeper is up upon initial setup",
                        default=True)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--poll', action='store_true', help="Polling example")
    group.add_argument('--event', action='store_true',
                       help="Event driven example")

    args = vars(parser.parse_args())

    hosts = args['zoo_keeper']
    file = args['file']
    pointer = args['pointer']
    sleep_time = args['sleep']
    poll = args['poll']
    event = args['event']
    valid = args['valid_init']

    def valid_file(state):
        print("The valid state is now", state)

    def change_file(conf_string):
        print("Your file contents:", conf_string)

    def error_file(message):
        print("An error was thrown:", message)

    # You can use any or all of these, polling + handlers, some handlers, etc
    if pointer:
        if poll:
            zoo_watcher = ZookeeperWatcher(hosts, file, pointer=True)
        elif event:
            zoo_watcher = ZookeeperWatcher(hosts, file,
                                           valid_handler=valid_file,
                                           config_handler=change_file,
                                           error_handler=error_file,
                                           pointer=True, valid_init=valid)
    else:
        if poll:
            zoo_watcher = ZookeeperWatcher(hosts, file)
        elif event:
            zoo_watcher = ZookeeperWatcher(hosts, file,
                                           valid_handler=valid_file,
                                           config_handler=change_file,
                                           error_handler=error_file,
                                           valid_init=valid)

    try:
        while True:
            if poll:
                print("Valid File?", zoo_watcher.is_valid())
                print("Contents:", zoo_watcher.get_file_contents())
            sleep(sleep_time)
    except:
        pass
    zoo_watcher.close()


if __name__ == "__main__":
    sys.exit(main())



'''
Offline tests
'''
from mock import MagicMock
from unittest import TestCase
from scrapy.http import Request
from crawling.redis_domain_max_page_filter import RFDomainMaxPageFilter
from crawling.redis_global_page_per_domain_filter import RFGlobalPagePerDomainFilter


class TestRedisPagePerDomainFilter(TestCase):

    def setUp(self):
        self.global_limit = RFGlobalPagePerDomainFilter(MagicMock(), 'key', 5, 600)
        self.per_domain_limit = RFDomainMaxPageFilter(MagicMock(), 'key', 600)

    def test_global_page_per_domain_filter(self):
        req = Request('http://example.com')
        req.meta['crawlid'] = "abc123"

        self.global_limit.server.expire = MagicMock()

        # successfully added
        self.global_limit.server.incr = MagicMock(return_value=4)
        self.assertFalse(self.global_limit.request_page_limit_reached(req, None))

        # unsuccessfully added
        self.global_limit.server.incr = MagicMock(return_value=5)
        self.assertTrue(self.global_limit.request_page_limit_reached(req, None))

    def test_domain_max_page_filter(self):
        req = Request('http://example.com')
        req.meta['crawlid'] = "abc123"
        req.meta['domain_max_pages'] = 5

        self.per_domain_limit.server.expire = MagicMock()

        # successfully added
        self.per_domain_limit.server.incr = MagicMock(return_value=4)
        self.assertFalse(self.per_domain_limit.request_page_limit_reached(req, None))

        # unsuccessfully added
        self.per_domain_limit.server.incr = MagicMock(return_value=6)
        self.assertTrue(self.per_domain_limit.request_page_limit_reached(req, None))



import tldextract
from scrapy.dupefilters import BaseDupeFilter
from scrapy.utils.reqser import request_to_dict


class RFGlobalPagePerDomainFilter(BaseDupeFilter):
    '''
    Redis-based request number filter
    When this filter is enabled, all crawl jobs have GLOBAL_PAGE_PER_DOMAIN_LIMIT as a hard limit
    of the max pages they are allowed to crawl for each individual spiderid+domain+crawlid combination.
    For example:
        spiderid = "spid_1"
        crawlid = "crawl_1"
        domains = ['domain1.test', 'domain2.test']
        GLOBAL_PAGE_PER_DOMAIN_LIMIT = 100
    then the crawler will fetch at most 100 pages of domain1.test and 100 pages of domain2.test.

    For each different spiderid or crawlid used, the domains will be crawled again for at most 100 pages each.

    Used when you don't want to pass domain_max_pages to all the individual CRAWL API requests.
    By default this filter is not enabled.
    '''

    def __init__(self, server, key, page_limit, timeout):
        '''
        Initialize page number filter
        @param server: the redis connection
        @param key: the key to store the fingerprints
        @param page_limit: the number of pages that when reached stops the particular domain crawl
        @param timeout: number of seconds a given key will remain once idle
        '''
        self.server = server
        # key_start equals: self.spider.name + ':global_page_count_filter'
        self.key_start = key
        self.page_limit = page_limit
        self.timeout = timeout
        # set up tldextract
        self.extract = tldextract.TLDExtract()

    def request_page_limit_reached(self, request, spider):
        # Collect items composing the redis key
        # grab the tld of the request
        req_dict = request_to_dict(request, spider)
        ex_res = self.extract(req_dict['url'])
        domain = "{d}.{s}".format(d=ex_res.domain, s=ex_res.suffix)

        # grab the crawl id
        crawl_id = req_dict['meta']['crawlid']

        # Compose the redis key
        composite_key = self.key_start + ':' + domain + ':' + crawl_id

        # Add new key if it doesn't exist
        if not self.server.exists(composite_key):
            self.server.set(composite_key, 0)

        # Stop incrementing the key when the limit is reached
        page_count = int(self.server.get(composite_key))
        if page_count >= self.page_limit:
            return True

        # Increment key
        page_count = int(self.server.incr(composite_key))
        # Set key expiration
        self.server.expire(composite_key, self.timeout)

        return page_count >= self.page_limit

    def close(self, reason):
        '''
        Delete data on close. Called by scrapy's scheduler
        '''
        self.clear()

    def clear(self):
        '''
        The page number per domain has a TTL so you shouldn't clear it
        '''
        pass



import argparse
from bhfutils.scutils.settings_wrapper import SettingsWrapper

# set up arg parser
parser = argparse.ArgumentParser(
    description='Example SettingsWrapper parser.\n')
parser.add_argument('-s', '--settings', action='store', required=False,
                    help="The default settings file",
                    default="settings.py")
parser.add_argument('-o', '--override-settings', action='store', required=False,
                    help="The override settings file",
                    default="localsettings.py")
parser.add_argument('-v', '--variable', action='store', required=False,
                    help="The variable to print out",
                    default=None)
args = vars(parser.parse_args())

# load up settings
wrapper = SettingsWrapper()
my_settings = wrapper.load(default=args['settings'],
                           local=args['override_settings'])

if args['variable'] is not None:
    if args['variable'] in my_settings:
        print(args['variable'], '=', my_settings[args['variable']])
    else:
        print(args['variable'], "not in loaded settings")
else:
    print("Full settings:", my_settings)


from builtins import object
#!/usr/bin/env python
# -*- coding: utf-8 -*-
import signal
"""
A timer class so methods can only execute for a certain amount of seconds,
then returns the default value

@author: Madison Bahmer @ IST Research
         Last Updated 9/1/15
"""


class MethodTimer(object):
    '''
    -------------------------------------
    Initial code from http://pguides.net/python-tutorial/python-timeout-a-function/
    Minor modifications made to work
    with classes, parameters, and new timeout name.

    Use above your function definition:
        @MethodTimer.timeout(seconds_to_wait, default_return_value)
        def myfunc(params):
    '''

    class DecoratorTimeout(Exception):
        '''
        Simple class in order to raise exception
        '''
        pass

    @staticmethod
    def timeout(timeout_time, default):
        '''
        Decorate a method so it is required to execute in a given time period,
        or return a default value.
        '''
        def timeout_function(f):
            def f2(*args):
                def timeout_handler(signum, frame):
                    raise MethodTimer.DecoratorTimeout()

                old_handler = signal.signal(signal.SIGALRM, timeout_handler)
                # triger alarm in timeout_time seconds
                signal.alarm(timeout_time)
                try:
                    retval = f(*args)
                except MethodTimer.DecoratorTimeout:
                    return default
                finally:
                    signal.signal(signal.SIGALRM, old_handler)
                signal.alarm(0)
                return retval
            return f2
        return timeout_function
    '''
    -------------------------------------
    '''
    def __init__(self):
        pass



'''
Online utils test
'''
from __future__ import division
from builtins import range
from past.utils import old_div
from builtins import object
import unittest
from unittest import TestCase
from mock import MagicMock
import time
import sys

import redis
import argparse
from kazoo.client import KazooClient

from bhfutils.scutils import RedisQueue, RedisPriorityQueue, RedisStack

from bhfutils.scutils import (ThreadedCounter, TimeWindow,
                              RollingTimeWindow, Counter, UniqueCounter,
                              HyperLogLogCounter, BitMapCounter)
from bhfutils.scutils import ZookeeperWatcher


class RedisMixin(object):

    def __init__(self, name, redis_conn):
        self.redis_conn = redis_conn
        TestCase.__init__(self, name)


class QueueMixin(RedisMixin, object):
    queue_class = None

    def setUp(self):
        self.queue = self.queue_class(self.redis_conn, 'key')


class TestRedisFifoQueue(QueueMixin, TestCase):
    queue_class = RedisQueue

    def test_fifo_queue(self):
        item1 = {"stuff": 'a', "blah": 2}
        item2 = 123
        item3 = "dohrayme"

        self.queue.push(item1)
        self.queue.push(item2)
        self.queue.push(item3)

        self.assertEqual(item1, self.queue.pop())
        self.assertEqual(item2, self.queue.pop())
        self.assertEqual(item3, self.queue.pop())


class TestRedisPriorityQueue(QueueMixin, TestCase):
    queue_class = RedisPriorityQueue

    def test_priority_queue(self):
        # Highest to lowest priority
        item1 = ['a', 'b', 'c']
        item2 = 123
        item3 = "dohrayme"

        self.queue.push(item1, 3)
        self.queue.push(item2, 1)
        self.queue.push(item3, 2)

        self.assertEqual(item1, self.queue.pop())
        self.assertEqual(item3, self.queue.pop())
        self.assertEqual(item2, self.queue.pop())


class TestRedisStack(QueueMixin, TestCase):
    queue_class = RedisStack

    def test_stack(self):
        item1 = ['a', 'b', 'c']
        item2 = 123
        item3 = "dohrayme"

        self.queue.push(item1)
        self.queue.push(item2)
        self.queue.push(item3)

        self.assertEqual(item3, self.queue.pop())
        self.assertEqual(item2, self.queue.pop())
        self.assertEqual(item1, self.queue.pop())


class TestStatsThreaded(RedisMixin, TestCase):

    def test_set_key(self):
        tc = ThreadedCounter(start_time=1442671176)
        self.assertEqual('default_counter', tc.get_key())

        tc.roll = True
        tc.key = 'myKey'
        tc._set_key()
        self.assertEqual('myKey:2015-09-19_13:59:36', tc.get_key())

    def test_is_expired(self):
        # test rolling the key we are using
        tc = ThreadedCounter(start_time=1442671176, window=5)
        tc._time = MagicMock(return_value=1442671177)
        self.assertFalse(tc.is_expired())

        tc._time = MagicMock(return_value=1442671182)
        self.assertTrue(tc.is_expired())

    def test_threading(self):
        # this test ensures the thread can start and stop
        tc = ThreadedCounter(start_time=1442671176, cycle_time=1)
        tc.expire = MagicMock()
        tc.setup(redis_conn=self.redis_conn)
        time.sleep(1)
        tc.stop()

    def test_purge_old(self):
        # test removing old keys
        tc = ThreadedCounter(keep_max=1)
        tc.keep_max = 1
        tc.redis_conn = self.redis_conn
        self.redis_conn.set('default_counter:2015-09', 'stuff')
        self.redis_conn.set('default_counter:2015-10', 'stuff2')

        tc.purge_old()
        self.assertEqual(['default_counter:2015-10'],
                         self.redis_conn.keys(tc.get_key() + ':*'))

        self.redis_conn.delete('default_counter:2015-10')

    def tearDown(self):
        keys = self.redis_conn.keys('default_counter:*')
        while len(keys) > 0:
            key = keys.pop()
            self.redis_conn.delete(key)


class CleanMixin(object):

    def clean_keys(self, key):
        keys = self.redis_conn.keys(key + '*')
        while len(keys) > 0:
            key = keys.pop()
            self.redis_conn.delete(key)


class TestStatsTimeWindow(RedisMixin, TestCase, CleanMixin):

    def test_window(self):
        counter = TimeWindow(key='test_key', cycle_time=.1, window=1)
        counter.setup(self.redis_conn)
        counter.increment()
        counter.increment()
        counter.increment()
        time.sleep(1.1)
        counter.increment()
        counter.stop()
        self.assertEqual(counter.value(), 3)

        self.clean_keys(counter.key)

    def test_roll_window(self):
        counter = TimeWindow(key='test_key', cycle_time=.01, window=1,
                             roll=True)
        counter.setup(self.redis_conn)
        counter.increment()
        counter.increment()
        counter.increment()
        time.sleep(1.5)
        counter.increment()
        counter.increment()
        counter.stop()

        keys = self.redis_conn.keys('test_key:*')
        keys.sort()
        self.assertEqual(len(keys), 2)
        self.assertEqual(self.redis_conn.zcard(keys[0]), 3)
        self.assertEqual(counter.value(), 2)
        counter.stop()
        self.clean_keys(counter.key)


class TestStatsRollingTimeWindow(RedisMixin, TestCase, CleanMixin):

    def test_rolling_window(self):
        # rough sleep to get us back on track
        time.sleep(5 - (int(time.time()) % 5))
        counter = RollingTimeWindow(key='test_key', cycle_time=.01, window=5)
        counter.setup(self.redis_conn)
        counter.increment()
        counter.increment()
        time.sleep(3)
        counter.increment()
        counter.increment()
        time.sleep(3)
        # at this point the first 2 counts have expired
        counter.increment()
        self.assertEqual(counter.value(), 3)
        time.sleep(5.1)
        # now everything has expired
        self.assertEqual(counter.value(), 0)
        # check to ensure counter is still working
        counter.increment()
        self.assertEqual(counter.value(), 1)
        counter.stop()
        self.clean_keys(counter.key)


class TestStatsCounter(RedisMixin, TestCase, CleanMixin):

    def test_generic_counter(self):
        counter = Counter(key='test_key')
        counter.setup(redis_conn=self.redis_conn)
        i = 0
        while i < 10000:
            counter.increment()
            i += 1

        self.assertEqual(10000, counter.value())
        counter.stop()
        self.clean_keys(counter.key)

    def test_roll_generic_counter(self):
        # rough sleep to get us back on track
        time.sleep(2 - (int(time.time()) % 2))
        counter = Counter(key='test_key', window=2, cycle_time=.1, roll=True)
        counter.setup(redis_conn=self.redis_conn)
        i = 0
        while i < 100:
            counter.increment()
            i += 1
        self.assertEqual(counter.value(), 100)
        time.sleep(3)
        # now the counter should have rolled
        i = 0
        while i < 50:
            counter.increment()
            i += 1
        self.assertEqual(counter.value(), 50)
        counter.stop()
        self.clean_keys(counter.key)


class TestStatsUniqueCounter(RedisMixin, TestCase, CleanMixin):

    def test_uniques(self):
        counter = UniqueCounter(key='test_key')
        counter.setup(redis_conn=self.redis_conn)

        i = 0
        while i < 1000:
            counter.increment(i % 100)
            i += 1
        self.assertEqual(counter.value(), 100)
        counter.stop()
        self.clean_keys(counter.key)

    def test_roll_uniques(self):
        # rough sleep to get us back on track
        time.sleep(2 - (int(time.time()) % 2))
        counter = UniqueCounter(key='test_key', window=2, cycle_time=.1,
                                roll=True)
        counter.setup(redis_conn=self.redis_conn)
        i = 0
        while i < 100:
            counter.increment(i % 10)
            i += 1
        self.assertEqual(counter.value(), 10)
        time.sleep(3)
        # now the counter should have rolled
        i = 0
        while i < 50:
            counter.increment(i % 10)
            i += 1
        self.assertEqual(counter.value(), 10)
        counter.stop()
        self.clean_keys(counter.key)


class TestStatsHyperLogLogCounter(RedisMixin, TestCase, CleanMixin):
    tolerance = 2  # percent

    def get_percent_diff(self, value, actual):
        return abs(actual - value) / (old_div((value + actual), 2.0)) * 100.0

    def test_hll_counter(self):
        counter = HyperLogLogCounter(key='test_key')
        counter.setup(redis_conn=self.redis_conn)
        for n in range(3):
            i = 0
            while i < 10010 * (n + 1):
                counter.increment(i % (5000 * (n + 1)))
                i += 1

            value = 5000 * (n + 1)
            actual = counter.value()
            diff = self.get_percent_diff(value, actual)
            self.assertLessEqual(diff, self.tolerance)
            self.clean_keys(counter.key)

        counter.stop()

    def test_roll_hll_counter(self):
        # rough sleep to get us back on track
        time.sleep(5.0 - (time.time() % 5.0))
        counter = HyperLogLogCounter(key='test_key', window=5, roll=True,
                                     cycle_time=.1, )
        counter.setup(redis_conn=self.redis_conn)
        i = 0
        while i < 5010:
            counter.increment(i % 1010)
            i += 1
        value = 1010
        actual = counter.value()
        diff = self.get_percent_diff(value, actual)
        self.assertLessEqual(diff, self.tolerance)
        self.clean_keys(counter.key)
        # rough sleep to get us back on track
        time.sleep(5.0 - (time.time() % 5.0))
        # we should be on to a new counter window by now
        i = 0
        while i < 5010:
            counter.increment(i % 3010)
            i += 1
        value = 3010
        actual = counter.value()
        diff = self.get_percent_diff(value, actual)
        self.assertLessEqual(diff, self.tolerance)
        counter.stop()
        self.clean_keys(counter.key)


class TestStatsBitMapCounter(RedisMixin, TestCase, CleanMixin):

    def test_bitmap_counter(self):
        counter = BitMapCounter(key='test_key')
        counter.setup(redis_conn=self.redis_conn)

        counter.increment(0)
        counter.increment(1)
        counter.increment(5)
        counter.increment(0)
        self.assertEqual(3, counter.value())
        counter.stop()
        self.clean_keys(counter.key)

    def test_roll_bitmap_counter(self):
        # rough sleep to get us back on track
        time.sleep(2 - (int(time.time()) % 2))
        counter = BitMapCounter(key='test_key', window=2, cycle_time=.1,
                                roll=True)
        counter.setup(redis_conn=self.redis_conn)

        counter.increment(0)
        counter.increment(1)
        counter.increment(0)
        self.assertEqual(2, counter.value())
        time.sleep(3)
        counter.increment(0)
        self.assertEqual(1, counter.value())
        counter.stop()
        self.clean_keys(counter.key)


class TestZookeeperWatcher(TestCase):
    def __init__(self, name, hosts):
        self.hosts = hosts
        self.file_path = '/test_path'
        self.file_data = 'test_data'
        self.pointer_path = '/test_pointer'
        self.pointer_data = self.file_path
        TestCase.__init__(self, name)

    def setUp(self):
        self.zoo_client = KazooClient(hosts=self.hosts)
        self.zoo_client.start()
        # prepare data
        self.zoo_client.ensure_path(self.file_path)
        self.zoo_client.set(self.file_path, self.file_data.encode('utf-8'))
        self.zoo_client.ensure_path(self.pointer_path)
        self.zoo_client.set(self.pointer_path,
                            self.pointer_data.encode('utf-8'))

        self.zoo_watcher = ZookeeperWatcher(hosts=self.hosts,
                                            filepath=self.file_path,
                                            pointer=False,
                                            ensure=False,
                                            valid_init=True)

    def test_ping(self):
        self.assertTrue(self.zoo_watcher.ping())

    def test_get_file_contents(self):
        pointer_zoo_watcher = ZookeeperWatcher(hosts=self.hosts,
                                               filepath=self.pointer_path,
                                               pointer=True,
                                               ensure=False,
                                               valid_init=True)

        self.assertEqual(self.zoo_watcher.get_file_contents(), self.file_data)
        self.assertEqual(pointer_zoo_watcher.get_file_contents(),
                         self.file_data)
        self.assertEqual(pointer_zoo_watcher.get_file_contents(True),
                         self.pointer_data)

        pointer_zoo_watcher.close()

    def tearDown(self):
        self.zoo_watcher.close()

        self.zoo_client.ensure_path(self.file_path)
        self.zoo_client.delete(self.file_path)
        self.zoo_client.ensure_path(self.pointer_path)
        self.zoo_client.delete(self.pointer_path)
        self.zoo_client.stop()
        self.zoo_client.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Online deployment Test"
                                                 " Script for Utils")
    parser.add_argument('-r', '--redis-host', action='store',
                        default='localhost', help="The Redis host ip")
    parser.add_argument('-p', '--redis-port', action='store', default='6379',
                        help="The Redis port")
    parser.add_argument('-P', '--redis-password', action='store', default=None,
                        help="The Redis password")
    parser.add_argument('-z', '--zoo-keeper', action='store',
                        default='localhost:2181',
                        help="The Zookeeper connection <host>:<port>")

    args = vars(parser.parse_args())
    redis_conn = redis.Redis(host=args['redis_host'], port=args['redis_port'],
                             password=args['redis_password'], decode_responses=True)

    # build testing suite
    suite = unittest.TestSuite()

    # moved to the top to help get better consistency
    suite.addTest(TestStatsHyperLogLogCounter('test_hll_counter', redis_conn))
    suite.addTest(TestStatsHyperLogLogCounter('test_roll_hll_counter',
                                              redis_conn))

    suite.addTest(TestRedisFifoQueue('test_fifo_queue', redis_conn))
    suite.addTest(TestRedisPriorityQueue('test_priority_queue', redis_conn))
    suite.addTest(TestRedisStack('test_stack', redis_conn))

    suite.addTest(TestStatsThreaded('test_set_key', redis_conn))
    suite.addTest(TestStatsThreaded('test_is_expired', redis_conn))
    suite.addTest(TestStatsThreaded('test_threading', redis_conn))
    suite.addTest(TestStatsThreaded('test_purge_old', redis_conn))

    suite.addTest(TestStatsTimeWindow('test_window', redis_conn))
    suite.addTest(TestStatsTimeWindow('test_roll_window', redis_conn))

    suite.addTest(TestStatsRollingTimeWindow('test_rolling_window',
                                             redis_conn))

    suite.addTest(TestStatsCounter('test_generic_counter', redis_conn))
    suite.addTest(TestStatsCounter('test_roll_generic_counter', redis_conn))

    suite.addTest(TestStatsUniqueCounter('test_uniques', redis_conn))
    suite.addTest(TestStatsUniqueCounter('test_roll_uniques', redis_conn))

    suite.addTest(TestStatsBitMapCounter('test_bitmap_counter', redis_conn))
    suite.addTest(TestStatsBitMapCounter('test_roll_bitmap_counter',
                                         redis_conn))
    suite.addTest(TestZookeeperWatcher('test_ping', args['zoo_keeper']))
    suite.addTest(TestZookeeperWatcher('test_get_file_contents',
                                       args['zoo_keeper']))

    result = unittest.TextTestRunner(verbosity=2).run(suite)

    if len(result.errors) > 0 or len(result.failures) > 0:
        sys.exit(1)
    else:
        sys.exit(0)



import argparse
from bhfutils.scutils.argparse_helper import ArgparseHelper

parser = argparse.ArgumentParser(
    description='example_ah.py: Prints various family members', add_help=False)
parser.add_argument('-h', '--help', action=ArgparseHelper,
                    help='show this help message and exit')
# use the default argparse setup, comment out the lines above
#parser = argparse.ArgumentParser(
#    description='example_ah.py: Prints various family members')

subparsers = parser.add_subparsers(help='commands', dest='command')

# args here are applied to all sub commands using the `parents` parameter
base_parser = argparse.ArgumentParser(add_help=False)
base_parser.add_argument('-n', '--name', action='store', required=True,
                        help="The name of the person running the program")

# subcommand 1, requires name of brother
bro_parser = subparsers.add_parser('bro', help='Prints only the brother\'s name',
                                   parents=[base_parser])
bro_parser.add_argument('-b', '--brother', action='store', required=True,
                         help="The brother's name")

# subcommand 2, requires name of sister and optionally mom
fam_parser = subparsers.add_parser('fam', help='Prints mom and sister\'s name',
                                   parents=[base_parser])
fam_parser.add_argument('-s', '--sister', action='store', required=True,
                         help="The sister's name")
fam_parser.add_argument('-m', '--mom', action='store', required=False,
                         default='Mom', help="The sister's name")

args = vars(parser.parse_args())

if args['command'] == 'bro':
    print("Your brother's name is " + args['brother'])
elif args['command'] == 'fam':
    print("Your sister's name is " + args['sister'] + " and you call your "\
        "Mother '" + args['mom'] + "'")


import mock
from scrapy import Item
from mock import MagicMock
from unittest import TestCase
from scrapy.http import Request
from crawling.meta_passthrough_middleware import MetaPassthroughMiddleware

class TestMetaPassthroughMiddleware(TestCase):

    @mock.patch('crawling.meta_passthrough_middleware' \
                '.MetaPassthroughMiddleware.setup')
    def setUp(self, s):
        self.mpm = MetaPassthroughMiddleware(MagicMock())
        self.mpm.logger = MagicMock()
        self.mpm.logger.debug = MagicMock()

    def test_mpm_middleware(self):
        # create fake response
        a = MagicMock()
        a.meta = {
            'key1': 'value1',
            'key2': 'value2'
        }

        yield_count = 0
        # test all types of results from a spider
        # dicts, items, or requests
        test_list = [
            {},
            Item(),
            Request('http://istresearch.com')
        ]

        for item in self.mpm.process_spider_output(a, test_list, MagicMock()):
            if isinstance(item, Request):
                self.assertEqual(a.meta, item.meta)
            yield_count += 1

        self.assertEqual(yield_count, 3)

        # 1 debug for the method, 1 debug for the request
        self.assertEqual(self.mpm.logger.debug.call_count, 2)

        # test meta unchanged if already exists
        r = Request('http://aol.com')
        r.meta['key1'] = 'othervalue'

        for item in self.mpm.process_spider_output(a, [r], MagicMock()):
            # key1 value1 did not pass through, since it was already set
            self.assertEqual(item.meta['key1'], 'othervalue')
            # key2 was not set, therefor it passed through
            self.assertEqual(item.meta['key2'], 'value2')



import sys
from os import path
from unittest import TestCase
from mock import MagicMock, patch
from kazoo.exceptions import KazooException
from bhfutils.scutils import ZookeeperWatcher

sys.path.append(path.dirname(path.dirname(path.abspath('utils'))))

class TestZookeeperWatcher(TestCase):

    def setUp(self):
        zoo_client = MagicMock()
        zoo_client.get = MagicMock(return_value=(b'data', 'blah'))

        with patch('utils.scutils.zookeeper_watcher.KazooClient') as k:
            k.return_value = zoo_client
            self.zoo_watcher = ZookeeperWatcher(
                hosts='localhost',
                filepath='/mypath',
                pointer=False, ensure=True,
                valid_init=True)

    def test_ping(self):
        self.zoo_watcher.zoo_client.server_version = MagicMock()
        self.assertTrue(self.zoo_watcher.ping())
        self.zoo_watcher.zoo_client.server_version = MagicMock(side_effect=KazooException)
        self.assertFalse(self.zoo_watcher.ping())

    def test_get_file_contents(self):
        self.zoo_watcher.old_pointed = 'old_pointed'
        self.zoo_watcher.old_data = 'old_data'

        self.zoo_watcher.pointer = False
        self.assertEqual(self.zoo_watcher.get_file_contents(), 'old_data')

        self.zoo_watcher.pointer = True
        self.assertEqual(self.zoo_watcher.get_file_contents(), 'old_data')

        self.zoo_watcher.pointer = True
        self.assertEqual(self.zoo_watcher.get_file_contents(True), 'old_pointed')

    def test_compare_pointer(self):
        self.zoo_watcher.old_pointed = '/path1'

        self.assertTrue(self.zoo_watcher.compare_pointer('/path2'))

        self.zoo_watcher.old_pointed = '/path1'

        self.assertFalse(self.zoo_watcher.compare_pointer('/path1'))

    def test_compare_data(self):
        self.zoo_watcher.old_data = 'old_data'

        self.assertTrue(self.zoo_watcher.compare_data('new_data'))

        self.zoo_watcher.old_data = 'same_data'
        self.assertFalse(self.zoo_watcher.compare_data('same_data'))

    def test_set_valid(self):
        self.zoo_watcher.is_valid = MagicMock(return_value=True)
        self.zoo_watcher.valid_handler = MagicMock()
        self.zoo_watcher.set_valid(False)

        self.zoo_watcher.valid_handler.assert_called_once_with(True)

    def test_call_valid(self):
        self.the_bool = False

        def the_set(state):
            self.the_bool = True

        self.zoo_watcher.valid_handler = the_set
        self.zoo_watcher.call_valid(True)

        self.assertTrue(self.the_bool)

    def test_call_config(self):
        self.the_bool = False

        def the_set(state):
            self.the_bool = True

        self.zoo_watcher.config_handler = the_set
        self.zoo_watcher.call_config(True)

        self.assertTrue(self.the_bool)

    def test_call_error(self):
        self.the_bool = False

        def the_set(state):
            self.the_bool = True

        self.zoo_watcher.error_handler = the_set
        self.zoo_watcher.call_error(True)

        self.assertTrue(self.the_bool)



from future import standard_library

standard_library.install_aliases()

try:
    import pickle as pickle
except ImportError:
    import pickle


class Base(object):
    '''
    Queue/Stack base class
    '''

    def __init__(self, server, key, encoding=pickle):
        '''Initialize the redis queue.

        @param server: the redis connection
        @param key: the key for this queue
        @param encoding: The encoding module to use.

        Note that if you wish to use any other encoding besides pickle, it
        is assumed you have already imported that module in your code before
        calling this constructor.
        '''
        self.server = server
        self.key = key
        self.encoding = encoding

        if not hasattr(self.encoding, 'dumps'):
            raise NotImplementedError("encoding does not support dumps()")
        if not hasattr(self.encoding, 'loads'):
            raise NotImplementedError("encoding does not support loads()")

    def _encode_item(self, item):
        '''
        Encode an item object

        @requires: The object be serializable
        '''
        if self.encoding.__name__ == 'pickle':
            return self.encoding.dumps(item, protocol=-1).decode('latin1')
        else:
            return self.encoding.dumps(item, reject_bytes=False)

    def _decode_item(self, encoded_item):
        '''
        Decode an item previously encoded
        '''
        if self.encoding.__name__ == 'pickle':
            return self.encoding.loads(encoded_item.encode('latin1'))
        else:
            return self.encoding.loads(encoded_item)

    def __len__(self):
        '''
        Return the length of the queue
        '''
        raise NotImplementedError

    def push(self, item):
        '''
        Push an item
        '''
        raise NotImplementedError

    def pop(self, timeout=0):
        '''
        Pop an item
        '''
        raise NotImplementedError

    def clear(self):
        '''
        Clear queue/stack
        '''
        self.server.delete(self.key)


class RedisQueue(Base):
    '''
    FIFO queue
    '''

    def __len__(self):
        '''
        Return the length of the queue
        '''
        return self.server.llen(self.key)

    def push(self, item):
        '''
        Push an item
        '''
        # ignore priority
        self.server.lpush(self.key, self._encode_item(item))

    def pop(self, timeout=0):
        '''
        Pop an item
        '''
        if timeout > 0:
            data = self.server.brpop(self.key, timeout)
            if isinstance(data, tuple):
                data = data[1]
        else:
            data = self.server.rpop(self.key)
        if data:
            return self._decode_item(data)


class RedisPriorityQueue(Base):
    '''
    Priority queue abstraction using redis' sorted set
    '''

    def __len__(self):
        '''Return the length of the queue'''
        return self.server.zcard(self.key)

    def push(self, item, priority):
        '''
        Push an item

        @param priority: the priority of the item
        '''
        data = self._encode_item(item)
        self.server.execute_command('ZADD', self.key, -priority, data)

    def pop(self, timeout=0):
        '''
        Pop an item
        timeout not support in this queue class
        '''
        # use atomic range/remove using multi/exec
        pipe = self.server.pipeline()
        pipe.multi()
        pipe.zrange(self.key, 0, 0).zremrangebyrank(self.key, 0, 0)
        results, count = pipe.execute()
        if results:
            return self._decode_item(results[0])


class RedisStack(Base):
    '''
    Stack
    '''

    def __len__(self):
        '''
        Return the length of the stack
        '''
        return self.server.llen(self.key)

    def push(self, item):
        '''
        Push an item
        '''
        self.server.lpush(self.key, self._encode_item(item))

    def pop(self, timeout=0):
        '''
        Pop an item
        '''
        if timeout > 0:
            data = self.server.blpop(self.key, timeout)
            if isinstance(data, tuple):
                data = data[1]
        else:
            data = self.server.lpop(self.key)

        if data:
            return self._decode_item(data)


__all__ = ['RedisQueue', 'RedisPriorityQueue', 'RedisStack']



'''
Offline tests
'''
from mock import MagicMock
from unittest import TestCase
from scrapy.http import Request
from scrapy.http import HtmlResponse
from crawling.items import RawResponseItem
from crawling.spiders.link_spider import LinkSpider


class TestLinkSpider(TestCase):

    def setUp(self):
        self.spider = LinkSpider()
        self.spider._logger = MagicMock()
        self.spider.stats_dict = {}

    def get_meta(self):
        item = {}
        item['crawlid'] = "abc123"
        item['appid'] = "myapp"
        item['spiderid'] = "link"
        item["attrs"] = {}
        item["allowed_domains"] = ()
        item["allow_regex"] = ()
        item["deny_regex"] = ()
        item["deny_extensions"] = None
        item['curdepth'] = 0
        item["maxdepth"] = 1
        item["domain_max_pages"] = None
        item['priority'] = 0
        item['retry_times'] = 0
        item['expires'] = 0

        return item

    def evaluate(self, meta_object,
                text, expected_raw, expected_requests):
        request = Request(url='http://www.drudgereport.com',
                          meta=meta_object)
        response = HtmlResponse('drudge.url', body=text, request=request,
                                encoding='utf8')

        raw_item_count = 0
        request_count = 0

        for x in self.spider.parse(response):
            if isinstance(x, RawResponseItem):
                raw_item_count = raw_item_count + 1
            elif isinstance(x, Request):
                request_count = request_count + 1

        self.assertEqual(raw_item_count, expected_raw)
        self.assertEqual(request_count, expected_requests)

    def test_link_spider_parse(self):
        with open('tests/drudge.html', 'r') as file:
            text = str(file.read())

            curr_meta = self.get_meta()

            # test too deep
            curr_meta['maxdepth'] = 1
            curr_meta['curdepth'] = 1
            self.evaluate(curr_meta, text, 1, 0)

            # test raw depth recursion
            curr_meta['curdepth'] = 0
            self.evaluate(curr_meta, text, 1, 391)

            # test allowed_domains filter
            curr_meta['allowed_domains'] = ['foxnews.com']
            self.evaluate(curr_meta, text, 1, 4)

            # test allow regex filter
            curr_meta['allowed_domains'] = ()
            curr_meta['allow_regex'] = ['.*shock.*']
            self.evaluate(curr_meta, text, 1, 1)

            # test deny regex filter
            curr_meta['allow_regex'] = ()
            curr_meta['deny_regex'] = ['.*.com.*']
            self.evaluate(curr_meta, text, 1, 43)

            # test deny_extensions filter
            curr_meta['deny_regex'] = ()
            # no pages that end in .html
            curr_meta['deny_extensions'] = ['html']
            self.evaluate(curr_meta, text, 1, 329)



import argparse
from getch import getch
from time import time
from bhfutils.scutils.stats_collector import StatsCollector

# set up arg parser
parser = argparse.ArgumentParser(
    description='Example key press stats collector.\n')
parser.add_argument('-rw', '--rolling-window', action='store_true',
                    required=False, help="Use a RollingTimeWindow counter",
                    default=False)
parser.add_argument('-r', '--redis-host', action='store', required=True,
                    help="The Redis host ip")
parser.add_argument('-p', '--redis-port', action='store', default='6379',
                    help="The Redis port")
parser.add_argument('-P', '--redis-password', action='store', default=None,
                    help="The Redis password")

args = vars(parser.parse_args())

the_window = StatsCollector.SECONDS_1_MINUTE

if args['rolling_window']:
    counter = StatsCollector.get_rolling_time_window(host=args['redis_host'],
                                                     port=args['redis_port'],
                                                     password=args['redis_password'],
                                                     window=the_window,
                                                     cycle_time=1)
else:
    counter = StatsCollector.get_time_window(host=args['redis_host'],
                                                     port=args['redis_port'],
                                                     password=args['redis_password'],
                                                     window=the_window,
                                                     keep_max=3)

print("Kill this program by pressing `ENTER` when done")

the_time = int(time())
floor_time = the_time % the_window
final_time = the_time - floor_time

pressed_enter = False
while not pressed_enter:
    print("The current counter value is " + str(counter.value()))
    key = getch()

    if key == '\r' or key == '\n':
        pressed_enter = True
    elif key == ' ':
        counter.increment()

    if not args['rolling_window']:
        new_time = int(time())
        floor_time = new_time % the_window
        new_final_time = new_time - floor_time

        if new_final_time != final_time:
            print("The counter window will roll soon")
            final_time = new_final_time

print("The final counter value is " + str(counter.value()))
counter.delete_key()



import argparse
from bhfutils.scutils.log_factory import LogFactory
parser = argparse.ArgumentParser(description='Example logger.')
parser.add_argument('-ll', '--log-level', action='store', required=False,
                    help="The log level", default='INFO',
                    choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'])
parser.add_argument('-lf', '--log-file', action='store_const',
                    required=False, const=False, default=True,
                    help='Log the output to the file. Otherwise logs to stdout')
parser.add_argument('-lj', '--log-json', action='store_const',
                    required=False, const=True, default=False,
                    help="Log the data in JSON format")
parser.add_argument('-ie', '--include-extra', action='store_const', const=True,
                        default=False, help="Print the 'extra' dict if not logging"
                        " to json")
args = vars(parser.parse_args())
logger = LogFactory.get_instance(level=args['log_level'], stdout=args['log_file'],
                    json=args['log_json'], include_extra=args['include_extra'])

my_var = 1

def the_callback(log_message, log_extras):
    global my_var
    my_var += 5

def the_callback_2(log_message, log_extras):
    global my_var
    my_var *= 2

logger.register_callback('DEBUG', the_callback)
logger.register_callback('WARN', the_callback_2, {'key':"value"})

logger.debug("debug output 1")
logger.warn("warn output", extra={"key":"value", "key2":"value2"})
logger.warn("warn output 2")
logger.debug("debug output 2")
logger.critical("critical fault, closing")
logger.debug("debug output 3")
sum = 2 + 2
logger.info("Info output closing.", extra={"sum":sum})
logger.error("Final var value", extra={"value": my_var})


import redis
import time
from builtins import object
from threading import Thread


class StatsCollector(object):
    '''
    A redis based statistics generator class. Use the following methods
    below to generate collectors for various statistic gathering.

    Example usage:

    from utils.stats_collector import StatsCollector
    # create a rolling time window that saves the last 24 hours of hits
    counter = StatsCollector.get_rolling_time_window(
            host='localhost',
            port=6379,
            password=None,
            window=StatsCollector.SECONDS_1_DAY)
    counter.increment()
    print counter.value()
    # that's it!
    '''
    # Easy to use time variables
    SECONDS_1_MINUTE = 60
    SECONDS_15_MINUTE = SECONDS_1_MINUTE * 15
    SECONDS_30_MINUTE = SECONDS_1_MINUTE * 30
    SECONDS_1_HOUR = SECONDS_1_MINUTE * 60
    SECONDS_2_HOUR = SECONDS_1_HOUR * 2
    SECONDS_4_HOUR = SECONDS_1_HOUR * 4
    SECONDS_6_HOUR = SECONDS_1_HOUR * 6
    SECONDS_12_HOUR = SECONDS_1_HOUR * 12
    SECONDS_24_HOUR = SECONDS_1_HOUR * 24
    SECONDS_48_HOUR = SECONDS_1_HOUR * 48
    SECONDS_1_DAY = SECONDS_24_HOUR
    SECONDS_2_DAY = SECONDS_1_DAY * 2
    SECONDS_3_DAY = SECONDS_1_DAY * 3
    SECONDS_7_DAY = SECONDS_1_DAY * 7
    SECONDS_1_WEEK = SECONDS_7_DAY
    SECONDS_30_DAY = SECONDS_1_DAY * 30
    REDIS_SOCKET_TIMEOUT = 10

    @classmethod
    def get_time_window(self, redis_conn=None, host='localhost', port=6379,
                        password=None, key='time_window_counter', cycle_time=5,
                        start_time=None, window=SECONDS_1_HOUR, roll=True,
                        keep_max=12):
        '''
        Generate a new TimeWindow
        Useful for collecting number of hits generated between certain times

        @param redis_conn: A premade redis connection (overrides host and port)
        @param host: the redis host
        @param port: the redis port
        @param password: the redis password
        @param key: the key for your stats collection
        @param cycle_time: how often to check for expiring counts
        @param start_time: the time to start valid collection
        @param window: how long to collect data for in seconds (if rolling)
        @param roll: Roll the window after it expires, to continue collecting
            on a new date based key.
        @keep_max: If rolling the static window, the max number of prior
            windows to keep
        '''
        counter = TimeWindow(key=key, cycle_time=cycle_time,
                             start_time=start_time, window=window, roll=roll,
                             keep_max=keep_max)
        counter.setup(redis_conn=redis_conn, host=host, port=port, password=password)
        return counter

    @classmethod
    def get_rolling_time_window(self, redis_conn=None, host='localhost',
                                port=6379, password=None,
                                key='rolling_time_window_counter',
                                cycle_time=5, window=SECONDS_1_HOUR):
        '''
        Generate a new RollingTimeWindow
        Useful for collect data about the number of hits in the past X seconds

        @param redis_conn: A premade redis connection (overrides host and port)
        @param host: the redis host
        @param port: the redis port
        @param password: the redis password
        @param key: the key for your stats collection
        @param cycle_time: how often to check for expiring counts
        @param window: the number of seconds behind now() to keep data for
        '''
        counter = RollingTimeWindow(key=key, cycle_time=cycle_time,
                                    window=window)
        counter.setup(redis_conn=redis_conn, host=host, port=port, password=password)
        return counter

    @classmethod
    def get_counter(self, redis_conn=None, host='localhost', port=6379,
                    password=None, key='counter', cycle_time=5, start_time=None,
                    window=SECONDS_1_HOUR, roll=True, keep_max=12, start_at=0):
        '''
        Generate a new Counter
        Useful for generic distributed counters

        @param redis_conn: A premade redis connection (overrides host and port)
        @param host: the redis host
        @param port: the redis port
        @param password: the redis password
        @param key: the key for your stats collection
        @param cycle_time: how often to check for expiring counts
        @param start_time: the time to start valid collection
        @param window: how long to collect data for in seconds (if rolling)
        @param roll: Roll the window after it expires, to continue collecting
            on a new date based key.
        @keep_max: If rolling the static window, the max number of prior
            windows to keep
        @param start_at: The integer to start counting at
        '''
        counter = Counter(key=key, cycle_time=cycle_time,
                          start_time=start_time, window=window, roll=roll,
                          keep_max=keep_max)
        counter.setup(redis_conn=redis_conn, host=host, port=port, password=password)
        return counter

    @classmethod
    def get_unique_counter(self, redis_conn=None, host='localhost', port=6379,
                           password=None, key='unique_counter', cycle_time=5,
                           start_time=None, window=SECONDS_1_HOUR, roll=True,
                           keep_max=12):
        '''
        Generate a new UniqueCounter.
        Useful for exactly counting unique objects

        @param redis_conn: A premade redis connection (overrides host and port)
        @param host: the redis host
        @param port: the redis port
        @param password: the redis password
        @param key: the key for your stats collection
        @param cycle_time: how often to check for expiring counts
        @param start_time: the time to start valid collection
        @param window: how long to collect data for in seconds (if rolling)
        @param roll: Roll the window after it expires, to continue collecting
            on a new date based key.
        @keep_max: If rolling the static window, the max number of prior
            windows to keep
        '''
        counter = UniqueCounter(key=key, cycle_time=cycle_time,
                                start_time=start_time, window=window,
                                roll=roll, keep_max=keep_max)
        counter.setup(redis_conn=redis_conn, host=host, port=port, password=password)
        return counter

    @classmethod
    def get_hll_counter(self, redis_conn=None, host='localhost', port=6379,
                        password=None, key='hyperloglog_counter', cycle_time=5,
                        start_time=None, window=SECONDS_1_HOUR, roll=True,
                        keep_max=12):
        '''
        Generate a new HyperLogLogCounter.
        Useful for approximating extremely large counts of unique items

        @param redis_conn: A premade redis connection (overrides host and port)
        @param host: the redis host
        @param port: the redis port
        @param password: the redis password
        @param key: the key for your stats collection
        @param cycle_time: how often to check for expiring counts
        @param start_time: the time to start valid collection
        @param window: how long to collect data for in seconds (if rolling)
        @param roll: Roll the window after it expires, to continue collecting
            on a new date based key.
        @keep_max: If rolling the static window, the max number of prior
            windows to keep
        '''
        counter = HyperLogLogCounter(key=key, cycle_time=cycle_time,
                                     start_time=start_time, window=window,
                                     roll=roll, keep_max=keep_max)
        counter.setup(redis_conn=redis_conn, host=host, port=port, password=password)
        return counter

    @classmethod
    def get_bitmap_counter(self, redis_conn=None, host='localhost', port=6379,
                           password=None, key='bitmap_counter', cycle_time=5,
                           start_time=None, window=SECONDS_1_HOUR, roll=True,
                           keep_max=12):
        '''
        Generate a new BitMapCounter
        Useful for creating different bitsets about users/items
        that have unique indices

        @param redis_conn: A premade redis connection (overrides host and port)
        @param host: the redis host
        @param port: the redis port
        @param password: the redis password
        @param key: the key for your stats collection
        @param cycle_time: how often to check for expiring counts
        @param start_time: the time to start valid collection
        @param window: how long to collect data for in seconds (if rolling)
        @param roll: Roll the window after it expires, to continue collecting
            on a new date based key.
        @keep_max: If rolling the static window, the max number of prior
            windows to keep
        '''
        counter = BitMapCounter(key=key, cycle_time=cycle_time,
                                start_time=start_time, window=window,
                                roll=roll, keep_max=keep_max)
        counter.setup(redis_conn=redis_conn, host=host, port=port, password=password)
        return counter


class AbstractCounter(object):

    def __init__(self, key=None):
        self.redis_conn = None
        if key is not None:
            self.key = key
        else:
            self.key = 'default_counter'

    def setup(self, redis_conn=None, host='localhost', port=6379, password=None):
        '''
        Set up the redis connection
        '''
        if redis_conn is None:
            if host is not None and port is not None:
                self.redis_conn = redis.Redis(host=host, port=port, password=password,
                                              decode_responses=True,
                                              socket_timeout=self.REDIS_SOCKET_TIMEOUT,
                                              socket_connect_timeout=self.REDIS_SOCKET_TIMEOUT)
            else:
                raise Exception("Please specify some form of connection to Redis")
        else:
            self.redis_conn = redis_conn

        self.redis_conn.info()

    def increment(self, **kwargs):
        '''
        Increments the counter by 1 if possible
        '''
        raise NotImplementedError("increment() method not implemented")

    def value(self, **kwargs):
        '''
        Returns the current count
        '''
        raise NotImplementedError("value() method not implemented")

    def expire(self, **kwargs):
        '''
        Expires items from the counter
        '''
        raise NotImplementedError("expire() method not implemented")

    def delete_key(self):
        '''
        Deletes the key being used
        '''
        self.redis_conn.delete(self.get_key())

    def _time(self):
        '''
        Returns the time
        '''
        return time.time()

    def get_key(self):
        '''
        Returns the key string
        '''
        return self.key


class ThreadedCounter(AbstractCounter):
    date_format = '%Y-%m-%d_%H:%M:%S'

    def __init__(self, key='default_counter', cycle_time=5, start_time=None,
                 window=None, roll=False, keep_max=5):
        '''
        A threaded counter, used to help roll time slots

        @param key: the key for your stats collection
        @param cycle_time: how often to check for expiring counts
        @param start_time: the time to start valid collection
        @param window: how long to collect data for in seconds (if rolling)
        @param roll: Roll the window after it expires, to continue collecting
            on a new date based key. Not applicable to the RollingWindow
        @keep_max: If rolling the static window, the max number of prior
            windows to keep
        '''
        AbstractCounter.__init__(self, key=key)

        if start_time is None and window is not None:
            # special case to auto generate correct start time
            the_time = int(self._time())
            floor_time = the_time % window
            self.start_time = the_time - floor_time
        else:
            self.start_time = start_time

        self.keep_max = None
        self.window = None
        self.cycle_time = cycle_time

        if window is not None:
            self.window = window
            self.roll = roll
            if self.roll and keep_max is not None and keep_max > 0:
                self.keep_max = keep_max
        else:
            self.roll = False

        self._set_key()

    def setup(self, redis_conn=None, host='localhost', port=6379, password=None):
        '''
        Set up the counting manager class

        @param redis_conn: A premade redis connection (overrides host and port)
        @param host: the redis host
        @param port: the redis port
        @param password: the redis password
        '''
        AbstractCounter.setup(self, redis_conn=redis_conn, host=host,
                              port=port, password=password)

        self._threaded_start()

    def _threaded_start(self):
        '''
        Spawns a worker thread to do the expiration checks
        '''
        self.active = True
        self.thread = Thread(target=self._main_loop)
        self.thread.setDaemon(True)
        self.thread.start()

    def deactivate(self):
        '''
        Call to shut down the threaded stats collector without joining; returns immediately
        '''
        self.active = False

    def stop(self):
        '''
        Call to shut down the threaded stats collector
        '''
        self.active = False
        self.thread.join()

    def _main_loop(self):
        """
        Main loop for the stats collector.
        Calls _do_thread_work that cleans up Redis sorted sets containing stats.
        """
        while self.active:
            self._do_thread_work()
        self._clean_up()

    def _do_thread_work(self):
        """
        Clean up Redis sorted sets containing stats.
        Runs inside a thread and communicates with Redis.
        Must be resilient to errors otherwise stats accumulate indefinitely.
        """
        try:  # Prevent thread from dying in case of Redis connectivity problems
            self.expire()  # Requires operational Redis connection
            if self.roll and self.is_expired():
                self.start_time = self.start_time + self.window
                self._set_key()
            self.purge_old()  # Requires operational Redis connection
        except redis.RedisError:
            pass
        finally:
            time.sleep(self.cycle_time)

    def _clean_up(self):
        '''
        Called after the main daemon thread is stopped
        '''
        pass

    def _set_key(self):
        '''
        sets the final key to be used currently
        '''
        if self.roll:
            self.date = time.strftime(self.date_format,
                                      time.gmtime(self.start_time))

            self.final_key = '{}:{}'.format(self.key, self.date)
        else:
            self.final_key = self.key

    def is_expired(self):
        '''
        Returns true if the time is beyond the window
        '''
        if self.window is not None:
            return (self._time() - self.start_time) >= self.window
        return False

    def purge_old(self):
        '''
        Removes keys that are beyond our keep_max limit
        '''
        if self.keep_max is not None:
            keys = self.redis_conn.keys(self.get_key() + ':*')
            keys.sort(reverse=True)
            while len(keys) > self.keep_max:
                key = keys.pop()
                self.redis_conn.delete(key)

    def get_key(self):
        return self.final_key


class TimeWindow(ThreadedCounter):

    def __init__(self, key='time_window_counter', cycle_time=5,
                 start_time=None, window=3600, roll=False, keep_max=None):
        '''
        A static window counter, is only valid for the initialized time range.

        @param start_time: the time to start valid collection
        @param window: how long to collect data for in seconds
        @param roll: Roll the window after it expires, to continue collecting
            on a new date based key
        '''
        ThreadedCounter.__init__(self, key=key, cycle_time=cycle_time,
                                 start_time=start_time, window=window,
                                 roll=roll, keep_max=keep_max)

    def increment(self):
        curr_time = self._time()
        if curr_time - self.start_time < self.window:
            self.redis_conn.zadd(self.final_key, {curr_time: curr_time})

    def value(self):
        return self.redis_conn.zcard(self.final_key)

    def expire(self):
        self.redis_conn.zremrangebyscore(self.final_key, '-inf',
                                         self.start_time - 1)
        self.redis_conn.zremrangebyscore(self.final_key,
                                         self.start_time + self.window, 'inf')


class RollingTimeWindow(ThreadedCounter):

    def __init__(self, key='rolling_time_window_counter', cycle_time=5,
                 window=60):
        '''
        A rolling time window. This continuously will have the number of hits
        within X seconds behind the current time.

        @param window: the collection window in seconds
        '''
        ThreadedCounter.__init__(self, key=key, cycle_time=cycle_time,
                                 start_time=None, window=window, roll=False,
                                 keep_max=None)
        self.window = window

    def increment(self):
        now = self._time()
        self.redis_conn.zadd(self.key, {now: now})

    def value(self):
        return self.redis_conn.zcard(self.key)

    def expire(self):
        expires = self._time() - self.window
        self.redis_conn.zremrangebyscore(self.key, '-inf', expires)


class Counter(ThreadedCounter):

    def __init__(self, key='counter', cycle_time=5, start_time=None,
                 window=None, roll=False, keep_max=None, start_at=0):
        '''
        A simple integer counter

        @param start_at: where to start the counter
        '''
        ThreadedCounter.__init__(self, key=key, cycle_time=cycle_time,
                                 start_time=start_time, window=window,
                                 roll=roll, keep_max=keep_max)

        self.init = False
        self.start_at = start_at

    def increment(self):
        if not self.init:
            self.redis_conn.set(self.final_key, self.start_at)
            self.init = True
        self.redis_conn.incr(self.final_key)

    def value(self):
        return int(self.redis_conn.get(self.final_key))

    def expire(self):
        pass


class UniqueCounter(ThreadedCounter):

    def __init__(self, key='unique_counter', cycle_time=5, start_time=None,
                 window=None, roll=False, keep_max=None):
        '''
        A unique item counter. Guarantees accuracy at the cost of storage
        '''
        ThreadedCounter.__init__(self, key=key, cycle_time=cycle_time,
                                 start_time=start_time, window=window,
                                 roll=roll, keep_max=keep_max)

    def increment(self, item):
        '''
        Tries to increment the counter by 1, if the item is unique

        @param item: the potentially unique item
        '''
        self.redis_conn.sadd(self.final_key, item)

    def value(self):
        return self.redis_conn.scard(self.final_key)

    def expire(self):
        pass


class HyperLogLogCounter(ThreadedCounter):

    def __init__(self, key='hyperloglog_counter', cycle_time=5,
                 start_time=None, window=None, roll=False, keep_max=None):
        '''
        A unique item counter. Accurate within 1%, max storage of 12k
        http://redis.io/topics/data-types-intro#hyperloglogs
        '''
        ThreadedCounter.__init__(self, key=key, cycle_time=cycle_time,
                                 start_time=start_time, window=window,
                                 roll=roll, keep_max=keep_max)

    def increment(self, item):
        '''
        Tries to increment the counter by 1, if the item is unique

        @param item: the potentially unique item
        '''
        self.redis_conn.execute_command("PFADD", self.final_key, str(item))

    def value(self):
        return self.redis_conn.execute_command("PFCOUNT", self.final_key)

    def expire(self):
        pass


class BitMapCounter(ThreadedCounter):

    def __init__(self, key='bitmap_counter', cycle_time=5, start_time=None,
                 window=None, roll=False, keep_max=None):
        '''
        A unique counter via bitmaps, see
        http://blog.getspool.com/2011/11/29/fast-easy-realtime-metrics-using-redis-bitmaps/
        for example usages
        '''
        ThreadedCounter.__init__(self, key=key, cycle_time=cycle_time,
                                 start_time=start_time, window=window,
                                 roll=roll, keep_max=keep_max)

    def increment(self, index):
        '''
        @param index: the index to set the flag on
        '''
        self.redis_conn.setbit(self.final_key, index, 1)

    def value(self):
        return self.redis_conn.execute_command("BITCOUNT", self.final_key)

    def expire(self):
        pass



'''
Offline utility tests
'''
import sys
import time
from os import path
from unittest import TestCase

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

from scutils.method_timer import MethodTimer


class TestMethodTimer(TestCase):

    def test_under(self):
        @MethodTimer.timeout(1, False)
        def method():
            time.sleep(0.5)
            return True
        result = method()
        self.assertTrue(result)

    def test_over(self):
        @MethodTimer.timeout(1, "STUFF")
        def method():
            time.sleep(1.5)
            return True
        result = method()
        self.assertEqual(result, "STUFF")

    def test_params(self):
        @MethodTimer.timeout(1, "STUFF2")
        def method(param1, param2, param3):
            time.sleep(1.5)
            return True
        result = method(True, "Stuff", ['item'])
        self.assertEqual(result, "STUFF2")



'''
Offline tests
'''
from mock import MagicMock
from unittest import TestCase
from scrapy.http import Request
from crawling.redis_retry_middleware import RedisRetryMiddleware


class TestRedisRetryMiddleware(TestCase):

    def setUp(self):
        self.retry = RedisRetryMiddleware(MagicMock())
        self.retry.max_retry_times = 1

    def test_retries(self):
        req = Request('http://example.com')
        req.meta['retry_times'] = 0
        req.meta['priority'] = 70

        # number of retries less than max
        out = self.retry._retry(req, 'stuff', MagicMock())
        self.assertEqual(req.url, out.url)
        self.assertEqual(60, out.meta['priority'])

        # over max
        self.assertEqual(self.retry._retry(out, 'stuff', MagicMock()), None)



from scrapy import signals
from scrapy.spiders import Spider
from scrapy.exceptions import DontCloseSpider


class RedisSpider(Spider):
    '''
    Base Spider for doing distributed crawls coordinated through Redis
    '''

    def _set_crawler(self, crawler):
        super(RedisSpider, self)._set_crawler(crawler)
        self.crawler.signals.connect(self.spider_idle,
                                     signal=signals.spider_idle)

    def spider_idle(self):
        raise DontCloseSpider

    def parse(self, response):
        '''
        Parse a page of html, and yield items into the item pipeline

        @param response: The response object of the scrape
        '''
        raise NotImplementedError("Please implement parse() for your spider")

    def set_logger(self, logger):
        '''
        Set the logger for the spider, different than the default Scrapy one

        @param logger: the logger from the scheduler
        '''
        self._logger = logger

    def reconstruct_headers(self, response):
        """
        Purpose of this method is to reconstruct the headers dictionary that
        is normally passed in with a "response" object from scrapy.

        Args:
            response: A scrapy response object

        Returns: A dictionary that mirrors the "response.headers" dictionary
        that is normally within a response object

        Raises: None
        Reason: Originally, there was a bug where the json.dumps() did not
        properly serialize the headers. This method is a way to circumvent
        the known issue
        """

        header_dict = {}
        # begin reconstructing headers from scratch...
        for key in list(response.headers.keys()):
            key_item_list = []
            key_list = response.headers.getlist(key)
            for item in key_list:
                key_item_list.append(item)
            header_dict[key] = key_item_list
        return header_dict



import tldextract
from scrapy.dupefilters import BaseDupeFilter
from scrapy.utils.reqser import request_to_dict


class RFDomainMaxPageFilter(BaseDupeFilter):
    '''
    Redis-based max page filter.
    This filter is applied per domain.
    Using this filter the maximum number of pages crawled
    for a particular domain is bounded.
    '''

    def __init__(self, server, key, timeout):
        '''
        Initialize page number filter

        @param server: the redis connection
        @param key: the key to store the fingerprints

        @param timeout: number of seconds a given key will remain once idle
        '''
        self.server = server
        # key_start equals: self.spider.name + ':domain_max_page_filter'
        self.key_start = key
        self.timeout = timeout
        # set up tldextract
        self.extract = tldextract.TLDExtract()

    def request_page_limit_reached(self, request, spider):
        # Collect items composing the redis key
        # grab the tld of the request
        req_dict = request_to_dict(request, spider)
        ex_res = self.extract(req_dict['url'])
        domain = "{d}.{s}".format(d=ex_res.domain, s=ex_res.suffix)

        # grab the crawl id
        crawl_id = req_dict['meta']['crawlid']

        # domain max page limit
        pagelimit = req_dict['meta']['domain_max_pages']

        # Compose the redis key
        composite_key = self.key_start + ':' + domain + ':' + crawl_id

        # Add new key if it doesn't exist
        if not self.server.exists(composite_key):
            self.server.set(composite_key, 0)

        # Stop incrementing the key when the limit is reached
        page_count = int(self.server.get(composite_key))
        if page_count >= pagelimit:
            # Prevent key expiration while the crawl continues by updating the expiration time
            self.server.expire(composite_key, self.timeout)
            return True

        # Increment key
        page_count = int(self.server.incr(composite_key))
        # Set key expiration
        self.server.expire(composite_key, self.timeout)

        return page_count >= pagelimit

    def close(self, reason):
        '''
        Delete data on close. Called by scrapy's scheduler
        '''
        self.clear()

    def clear(self):
        '''
        The page number per domain has a TTL so you shouldn't clear it
        '''
        pass



from scrapy.downloadermiddlewares.cookies import CookiesMiddleware


class CustomCookiesMiddleware(CookiesMiddleware):
    '''
    Custom Cookies Middleware to pass our required cookies along but not
    persist between calls
    '''

    def process_request(self, request, spider):
        if 'dont_merge_cookies' in request.meta:
            return

        cookiejarkey = request.meta.get("cookiejar")
        jar = self.jars[cookiejarkey]
        jar.clear()  # custom line
        cookies = self._get_request_cookies(jar, request)
        for cookie in cookies:
            jar.set_cookie_if_ok(cookie, request)

        # set Cookie header
        request.headers.pop('Cookie', None)
        jar.add_cookie_header(request)

        self._debug_cookie(request, spider)



import argparse
from time import sleep
from bhfutils.scutils.zookeeper_watcher import ZookeeperWatcher

parser = argparse.ArgumentParser(
    description="Zookeeper file watcher")
parser.add_argument('-z', '--zoo-keeper', action='store', required=True,
                    help="The Zookeeper connection <host>:<port>")
parser.add_argument('-f', '--file', action='store', required=True,
                    help="The full path to the file to watch in Zookeeper")
parser.add_argument('-p', '--pointer', action='store_const', const=True,
                    help="The file contents point to another file")
parser.add_argument('-s', '--sleep', nargs='?', const=1, default=1,
                    type=int, help="The time to sleep between poll checks")
parser.add_argument('-v', '--valid-init', action='store_false',
                    help="Do not ensure zookeeper is up upon initial setup",
                    default=True)
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('--poll', action='store_true', help="Polling example")
group.add_argument('--event', action='store_true',
                   help="Event driven example")

args = vars(parser.parse_args())

hosts = args['zoo_keeper']
file = args['file']
pointer = args['pointer']
sleep_time = args['sleep']
poll = args['poll']
event = args['event']
valid = args['valid_init']


def valid_file(state):
    print("The valid state is now", state)


def change_file(conf_string):
    print("Your file contents:", conf_string)


def error_file(message):
    print("An error was thrown:", message)


# You can use any or all of these, polling + handlers, some handlers, etc
if pointer:
    if poll:
        zoo_watcher = ZookeeperWatcher(hosts, file, ensure=True, pointer=True)
    elif event:
        zoo_watcher = ZookeeperWatcher(hosts, file,
                                       valid_handler=valid_file,
                                       config_handler=change_file,
                                       error_handler=error_file,
                                       pointer=True, ensure=True, valid_init=valid)
else:
    if poll:
        zoo_watcher = ZookeeperWatcher(hosts, file, ensure=True)
    elif event:
        zoo_watcher = ZookeeperWatcher(hosts, file,
                                       valid_handler=valid_file,
                                       config_handler=change_file,
                                       error_handler=error_file,
                                       valid_init=valid, ensure=True)

print("Use a keyboard interrupt to shut down the process.")
try:
    while True:
        if poll:
            print("Valid File?", zoo_watcher.is_valid())
            print("Contents:", zoo_watcher.get_file_contents())
        sleep(sleep_time)
except:
    pass
zoo_watcher.close()



from scrapy.downloadermiddlewares.retry import RetryMiddleware


unserializable_meta_params = ['playwright_page_methods']


class RedisRetryMiddleware(RetryMiddleware):

    def __init__(self, settings):
        RetryMiddleware.__init__(self, settings)

    def _retry(self, request, reason, spider):
        retries = request.meta.get('retry_times', 0) + 1
        if retries <= self.max_retry_times:
            retryreq = request.copy()
            # remove unserializable parameters
            for unserializable_meta_param in unserializable_meta_params:
                if unserializable_meta_param in retryreq.meta:
                    del retryreq.meta[unserializable_meta_param]
            # add params
            retryreq.meta['retry_times'] = retries
            retryreq.dont_filter = True
            # low priority
            retryreq.meta['priority'] = retryreq.meta['priority'] - 10

            return retryreq



'''
Offline utility tests
'''
import sys
import argparse
from os import path
from mock import MagicMock
from unittest import TestCase

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

from scutils.argparse_helper import ArgparseHelper

# from http://stackoverflow.com/questions/4219717/how-to-assert-output-with-nosetest-unittest-in-python
from contextlib import contextmanager
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class TestArgparseHelper(TestCase):

    def test_output(self):
        parser = argparse.ArgumentParser(description='Desc here',
        								 add_help=False)
        parser.add_argument('-h', '--help', action=ArgparseHelper, help='show this help message and exit')

        subparsers = parser.add_subparsers(help='commands', dest='command')

        base_parser = argparse.ArgumentParser(add_help=False)
        base_parser.add_argument('-s', '--settings', action='store',
                                 required=False,
                                 help="The settings file to read from",
                                 default="localsettings.py")

        feed_parser = subparsers.add_parser('feed', help='Feed the script',
                                            parents=[base_parser])
        feed_parser.add_argument('json', help='The JSON object as a string')

        run_parser = subparsers.add_parser('run', help='Run the script',
                                           parents=[base_parser])

        a = ArgparseHelper(MagicMock())

        expected = '''usage: nosetests [-h] {feed,run} ...

Desc here

positional arguments:
  {feed,run}  commands
    feed      Feed the script
    run       Run the script

optional arguments:
  -h, --help  show this help message and exit

Command 'feed'
usage: nosetests feed [-h] [-s SETTINGS] json

Command 'run'
usage: nosetests run [-h] [-s SETTINGS]'''

        try:
            with captured_output() as (out, err):
                a(parser, MagicMock(), MagicMock())
        except SystemExit:
            pass

        output = out.getvalue().strip()
        self.assertEqual(output, expected)
